#include <stdio.h>
#include <stdlib.h>
#include <i86.h>
#include <dos.h>

#define USE_IRQ

extern void __far * __based(__segname("_CODE")) old_int13_hdl;

int dos_close(unsigned handle);
#pragma aux dos_close = \
  "mov ah, 0x3e" \
  "int 0x21" \
  parm [bx] value [ax];

void outbyte(unsigned port, unsigned char byte);
#pragma aux outbyte = \
  "out dx, al" \
  parm [dx] [al];

unsigned char inbyte(unsigned port);
#pragma aux inbyte = \
  "in al, dx" \
  parm [dx] value [al];

void in_words(unsigned port, unsigned count, void __far *buff);
#ifdef __SW_0
#pragma aux in_words = \
  "next:" \
  "in ax, dx" \
  "mov es:[di], ax" \
  "add di, 2" \
  "loop next" \
  parm [dx] [cx] [es di] modify [ax dx cx di];
#else
#pragma aux in_words = \
  "push ds" \
  "push es" \
  "pop ds" \
  "cld" \
  "rep insw" \
  "pop ds" \
  parm [dx] [cx] [es di];
#endif
  

void out_words(unsigned port, unsigned count, void __far *buff);
#ifdef __SW_0
#pragma aux out_words = \
  "next:" \
  "mov ax, es:[di]" \
  "out dx, ax" \
  "add di, 2" \
  "loop next" \
  parm [dx] [cx] [es di] modify [ax dx cx di];
#else
#pragma aux out_words = \
  "push ds" \
  "push es" \
  "pop ds" \
  "cld" \
  "rep outsw" \
  "pop ds" \
  parm [dx] [cx] [es si];
#endif

void disable(void);
#pragma aux disable = \
  "cli";
void enable(void);
#pragma aux enable = \
  "sti";

unsigned long clock_ticks(void);
#pragma aux clock_ticks = \
  "xor ah, ah" \
  "int 0x1a" \
  value [cx dx];

void halt(void);
#pragma aux halt = \
  "hlt";

// wait 400 ns
void atapi_delay(unsigned port);
#pragma aux atapi_delay = \
  "add dx, 0x206" \
  "mov cx, 4" \
  "next:" \
  "in al, dx" \
  "loop next" \
  parm [dx] modify [ax cx dx];


unsigned dev_port(unsigned char devnum)
{
  if(devnum < 2) return 0x1F0;
  if(devnum < 4) return 0x170;
  return 0;
}

unsigned char bios_error_code(unsigned port)
{
  unsigned char error = inbyte(port + 1);
  if(error & 0x01) return 2;
  if(error & 0x02) return 0x0C;
  if(error & 0x04) return 1;
  if(error & 0x08) return 6;
  if(error & 0x10) return 2;
  if(error & 0x20) return 6;
  if(error & 0x40) return 2;
  if(error & 0x80) return 2;
  return 0;
}

#define error_code *(unsigned char __far*)MK_FP(0x40, 0x41)

int wait_device_busy(unsigned port)
{
  unsigned long time = clock_ticks();
  unsigned char status;

  while(1) // loop while the device is busy
  {
    status = inbyte(port + 7);
    if(!(status & 0x80)) return 0;
    if(clock_ticks() > time + 14) return 0x80;
  }
}

int wait_device_ready(unsigned port)
{
  unsigned long time = clock_ticks();
  while(1)
  {
    if(inbyte(port + 7) & 0x40) return 0;
    if(clock_ticks() > time + 14) return 0x80;
  }
}

int wait_data_request(unsigned port)
{
  unsigned long time = clock_ticks();
  unsigned char status;

  while(1) // loop until DRQ is set
  {
    status = inbyte(port + 7);
    if(!(status & 0x80))
    {
      if(status & 0x01) return bios_error_code(port);
      if(status & 0x08)
      {
        atapi_delay(port);
        return 0;
      }
      return 0x20;
    }
    if(clock_ticks() > time + 5*14) return 0x80;
  }
}

int drive_select(unsigned port, unsigned char devnum)
{
  unsigned char slave = devnum & 1;
  outbyte(port + 6, 0xE | (slave << 4));
  atapi_delay(port);
  return wait_device_busy(port);
}

int identify_packet_device(unsigned char devnum)
{
  unsigned char buffer[512];
  unsigned port = dev_port(devnum);
  int rc;

  if(rc = wait_device_busy(port)) return rc;
  if(rc = drive_select(port, devnum)) return rc;
  outbyte(port + 7, 0xA1); // send IDENTIFY PACKET DEVICE command
  atapi_delay(port);
  if(rc = wait_data_request(port)) return rc;
  in_words(port, 256, buffer); // read data from the device

  if(((*(unsigned*)&buffer[0]) & 0xDF03) == 0x8500)
    return 0;
  return 0xAA;
}


void __far *get_vector(unsigned char intr);
#pragma aux get_vector = \
  "mov ah, 0x35" \
  "int 0x21" \
  parm [al] value [es bx];

void set_vector(unsigned char intr, void __far *ptr);
#pragma aux set_vector = \
  "push ds"  \
  "push es" \
  "pop ds" \
  "mov ah, 0x25" \
  "int 0x21" \
  "pop ds" \
  parm [al] [es dx];


volatile unsigned char __based(__segname("_CODE")) irq;
void __far * __based(__segname("_CODE")) old_ata_int_hdl;
unsigned char __based(__segname("_CODE")) irq_num;
unsigned char __based(__segname("_CODE")) old_irq_mask;

unsigned bios_idle(void);
#pragma aux bios_idle = \
  "mov ax, 0x9000" \
  "int 0x15" \
  "lahf" \
  value [ax];

void bios_int_complete(void);
#pragma aux bios_int_complete = \
  "mov ax, 0x9100" \
  "int 0x15" \
  modify [ax];


void __interrupt ata_int_hdl(void)
{
  irq = 1;
  outbyte(0x20, 0x20);
  outbyte(0xA0, 0x20);
  bios_int_complete();
}

int wait_irq(unsigned port)
{
  unsigned long time = clock_ticks();

  if(bios_idle() & 0x0100)
    return 0x80;

  while(1)
  {
    if(irq || clock_ticks() > time + 14)
    {
      return wait_data_request(port);
    }
  }
}

void set_irq_vector(unsigned devnum)
{
  if(devnum < 2)
    irq_num = 0x76;
  else
    irq_num = 0x77;
  old_ata_int_hdl = get_vector(irq_num);
  set_vector(irq_num, ata_int_hdl);
  old_irq_mask = inbyte(0xA1);
  outbyte(0xA1, old_irq_mask & ~(0x80 > (devnum > 1)));
}

void unset_irq_vector(void)
{
  set_vector(irq_num, old_ata_int_hdl);
  outbyte(0xA1, old_irq_mask);
}

int send_packet_cmd(unsigned port, unsigned char __far *cmd, unsigned len)
{
  int rc;
  outbyte(port+7, 0xA0); // PACKET command
  atapi_delay(port);
  if(rc = wait_data_request(port)) return rc;
  out_words(port, len>>1, cmd);
  return 0;
}

unsigned transfer_size(unsigned port)
{
  return ((unsigned)inbyte(port + 5) << 8) | inbyte(port + 4);
}

int read_sectors(unsigned char devnum, unsigned long sector, unsigned char __far *count, unsigned char __far *buff)
{
  unsigned char read_cmd[12];
  unsigned port = dev_port(devnum);
  unsigned block_size;
  int rc;

  read_cmd[0] = 0xA8;
  read_cmd[1] = 0x00;
  read_cmd[2] = sector >> 24;
  read_cmd[3] = sector >> 16;
  read_cmd[4] = sector >> 8;
  read_cmd[5] = sector;
  read_cmd[6] = 0;
  read_cmd[7] = 0;
  read_cmd[8] = 0;
  read_cmd[9] = *count;
  read_cmd[10] = 0;
  read_cmd[11] = 0;

  if(rc = wait_device_busy(port)) return rc;
  if(rc = drive_select(port, devnum)) return rc;
  outbyte(port + 1, 0); // features register = 0 (PIO transfer)
  outbyte(port + 4, 0); // lba mid register = 0x00
  outbyte(port + 5, 2048 >> 8); // lba high register = 0x08
  atapi_delay(port);
  if(rc = wait_device_ready(port)) return rc;

#ifdef USE_IRQ
  set_irq_vector(devnum);
  irq = 0;
#endif

  if(rc = send_packet_cmd(port, read_cmd, 12)) return rc; // send READ(12) command

  while(*count)
  {
#ifdef USE_IRQ
    if(!irq)
    {
      if(rc = wait_irq(port)) break;
    }
    else
    {
#endif
    atapi_delay(port);
    if(rc = wait_data_request(port)) break;
#ifdef USE_IRQ
    }
#endif
    block_size = transfer_size(port);
/*#ifdef USE_IRQ
    irq = 0;
#endif*/
    in_words(port, block_size>>1, buff);
    buff += block_size;
    --*count;
  }

#ifdef USE_IRQ
  unset_irq_vector();
#endif
  return rc;
}

int lba_read(unsigned char devnum, unsigned char __far *lba_struct)
{
  unsigned long sector;
  unsigned char count;
  void __far *buff;
  int rc;
  if(lba_struct[0] < 16) return 1;
  count = lba_struct[2];
  buff = *(void __far* __far*)&lba_struct[4];
  sector = *(unsigned long __far*)&lba_struct[8];
  rc = read_sectors(devnum, sector, &count, buff);
  lba_struct[2] -= count;
  return rc;
}

int device_parameters(unsigned char const devnum, unsigned char __far * const parameter_struct)
{
  unsigned buffer_len = *(unsigned __far*)&parameter_struct[0];
  if(buffer_len < 26) return 1;
  if(buffer_len >= 30) *(unsigned long __far*)&parameter_struct[26] = -1;
  *(unsigned __far*)&parameter_struct[2] = 0x0004;
  return 0;
}

void __interrupt int13h_handler(union INTPACK regs)
{
  const unsigned char function = regs.h.ah;
  const unsigned char device = regs.h.dl;
  unsigned char __far * const buffer = MK_FP(regs.w.ds, regs.w.si);
  const unsigned char devnum = device & 0x0F;
  int rc;

  enable();

  switch(function)
  {
    case 0x41: // check extensions present
      if(regs.w.bx == 0x55AA && !(rc = identify_packet_device(devnum)))
      {
        regs.w.bx = 0xAA55;
        regs.h.ah = 0x21;
        regs.w.cx = 0x01;
        regs.w.flags &= ~1;
        error_code = 0;
      }
      else
      {
        error:
        regs.w.flags |= 1;
        regs.h.ah = error_code = rc;
      }
      return;
    case 0x42: // extended read
      if(!(rc = lba_read(devnum, buffer)))
      {
        regs.w.ax = error_code = 0;
        regs.w.flags &= ~1;
        return;
      }
      else
      {
        goto error;
      }
    case 0x48: // get device parameters
      if((rc = device_parameters(devnum, buffer)))
        goto error;
      regs.w.ax = error_code = 0;
      regs.w.flags &= ~1;
      return;
    default:
      regs.w.ax = 1;
      regs.w.flags |= 1;
      return;
  }
}
/*
int send_inquiry_cmd(unsigned char devnum, unsigned char __far *alloc_len, unsigned char *buffer)
{
  unsigned port = dev_port(devnum);
  unsigned char inquiry_cmd[6];  
  int rc;
  unsigned char n;

  inquiry_cmd[0] = 0x12;
  inquiry_cmd[1] = 0;
  inquiry_cmd[2] = 0;
  inquiry_cmd[3] = 0;
  inquiry_cmd[4] = *alloc_len;
  inquiry_cmd[5] = 0;

  if(rc = wait_device_busy(port)) return rc;
  if(rc = drive_select(port, devnum)) return rc;
  if(rc = wait_device_ready(port)) return rc;

  if(rc = send_packet_cmd(port, inquiry_cmd, 6)) return rc;

  atapi_delay(port);

  if(rc = wait_data_request(port)) return rc;

  *alloc_len = n = transfer_size(port);
  printf("%u\n", n);
  if(n)
  {
    if(*alloc_len & 1) ++*alloc_len;
    if(*alloc_len > 36) *alloc_len = 36;

    in_words(port, *alloc_len>>1, buffer);
    return 0;
  }
  else return 1;
}

int inquiry(unsigned char devnum)
{
  int rc;
  unsigned char buffer[36];
  unsigned char alloc_len = 5;

  if(devnum < 2) printf("Primary ");
  else printf("Secondary ");
  if(devnum&1) printf("slave: ");
  else printf("master: ");

  rc = send_inquiry_cmd(devnum, &alloc_len, buffer);
  if(!rc && alloc_len < 36)
    rc = send_inquiry_cmd(devnum, &buffer[4]+4, buffer);

  if(!rc && alloc_len)
    printf("Make: %.8s Model: %.16s\n", &buffer[8], &buffer[16]);

  return rc;
}

void detect_drives(void)
{
  unsigned n;
  puts("Detecting ATA devices...");
  for(n=0;n<4;n++)
    if(inquiry(n)) puts("Not detected");
}
*/
void __far int13h_entry();

unsigned current_psp();
#pragma aux current_psp = \
  "mov ah, 0x51" \
  "int 0x21" \
  value [bx] modify [ax];

void dos_free(unsigned);
#pragma aux dos_free = \
  "mov ah, 0x49" \
  "mov es, bx" \
  "int 0x21" \
  parm [bx] modify [es ax];

#pragma aux main aborts;
void main()
{
  unsigned char __far *psp_ptr = MK_FP(current_psp(), 0);
  unsigned env_seg;
  unsigned n;

  old_int13_hdl = get_vector(0x13);
  set_vector(0x13, int13h_entry);

  puts("BIOS int13h extensions for ATAPI CD-ROM drives installed.");
  env_seg = *(unsigned __far*)&psp_ptr[0x2C];
  dos_free(env_seg);

  // close stdio
  for(n=3;n--;)
    dos_close(n);

  _dos_keep(0, 0x10 + ((FP_OFF(main) + 0xF) >> 4));
}
