.8086

.MODEL tiny

LEFT_BUTTON EQU 0x01
CENTER_BUTTON EQU 0x04
RIGHT_BUTTON EQU 0x02
MOUSE_ENABLED EQU 0x08

DIR_LEFT EQU 0x10
DIR_RIGHT EQU 0x20
DIR_UP EQU 0x40
DIR_DOWN EQU 0x80

KEYPAD_LEFT EQU 0x4B
KEYPAD_RIGHT EQU 0x4D
KEYPAD_UP EQU 0x48
KEYPAD_DOWN EQU 0x50

KEY_CTRL EQU 0x1D
KEY_ENABLEMOUSE EQU KEY_CTRL

KEYPAD_INS EQU 0x52
MOUSE_LEFT EQU KEYPAD_INS

KEYPAD_ENTER EQU 0x1C
MOUSE_RIGHT EQU KEYPAD_ENTER

KEYPAD_PLUS EQU 0x4e
MOUSE_CENTER EQU KEYPAD_PLUS


.code

ORG 0x100

start:

mov ax, offset code_start
push ax
jmp init_driver_

; global variables
mouse_button_status: db 0
mouse_x_speed: dw 0
mouse_y_speed: dw 0
cursor_row: dw 100
cursor_col: dw 160
mouse_dx: dw 0
mouse_dy: dw 0

ax_save0: dw 0

accelerate_left:
  dec word ptr cs:mouse_x_speed
  x_addition:
  mov cs:ax_save0, ax
  mov ax, cs:mouse_x_speed
  add word ptr cs:cursor_col, ax
  add word ptr cs:mouse_dx, ax
  mov ax, cs:ax_save0
  ret

accelerate_right:
  inc word ptr cs:mouse_x_speed
  jmp x_addition

accelerate_up:
  dec word ptr cs:mouse_y_speed
  y_addition:
  mov cs:ax_save0, ax
  mov ax, cs:mouse_y_speed
  add word ptr cs:cursor_row, ax
  add word ptr cs:mouse_dy, ax
  mov ax, cs:ax_save0
  ret

accelerate_down:
  inc word ptr cs:mouse_y_speed
  jmp y_addition

ax_save1: dw 0

keyboard_handler:
  mov cs:ax_save1, ax

  in al, 0x60

  cmp al, KEY_ENABLEMOUSE
  jne no_enable_key
  or cs:mouse_button_status, MOUSE_ENABLED
  jmp keyboard_handler_chain

  no_enable_key:
  test cs:mouse_button_status, MOUSE_ENABLED
  jz keyboard_handler_chain

  cmp al, KEYPAD_LEFT
  jne no_left
  test byte ptr cs:mouse_button_status, DIR_LEFT
  jnz keyboard_handler_end
  or byte ptr cs:mouse_button_status, DIR_LEFT
  call accelerate_left
  jmp keyboard_handler_end

  no_left:
  cmp al, KEYPAD_RIGHT
  jne no_right
  test byte ptr cs:mouse_button_status, DIR_RIGHT
  jnz keyboard_handler_end
  or byte ptr cs:mouse_button_status, DIR_RIGHT
  call accelerate_right
  jmp keyboard_handler_end

  no_right:
  cmp al, KEYPAD_DOWN
  jne no_down
  test byte ptr cs:mouse_button_status, DIR_DOWN
  jnz keyboard_handler_end
  or byte ptr cs:mouse_button_status, DIR_DOWN
  call accelerate_down
  jmp keyboard_handler_end

  no_down:
  cmp al, KEYPAD_UP
  jne no_up
  test byte ptr cs:mouse_button_status, DIR_UP
  jnz keyboard_handler_end
  or byte ptr cs:mouse_button_status, DIR_UP
  call accelerate_up
  jmp keyboard_handler_end

  no_up:
  cmp al, MOUSE_LEFT
  jne no_mouseleft
  or byte ptr cs:mouse_button_status, LEFT_BUTTON
  jmp keyboard_handler_end

  no_mouseleft:
  cmp al, MOUSE_RIGHT
  jne no_mouseright
  or byte ptr cs:mouse_button_status, RIGHT_BUTTON
  jmp keyboard_handler_end

  no_mouseright:
  cmp al, MOUSE_CENTER
  jne no_mousecenter
  or byte ptr cs:mouse_button_status, CENTER_BUTTON
  jmp keyboard_handler_end

  no_mousecenter:

  and al, 0x7F

  cmp al, KEYPAD_LEFT
  je stop_hor
  cmp al, KEYPAD_RIGHT
  je stop_hor
  cmp al, KEYPAD_DOWN
  je stop_ver
  cmp al, KEYPAD_UP
  je stop_ver
  cmp al, MOUSE_LEFT
  jne nomouseleftup
  and byte ptr cs:mouse_button_status, NOT LEFT_BUTTON
  nomouseleftup:
  cmp al, MOUSE_RIGHT
  jne nomouserightup
  and byte ptr cs:mouse_button_status, NOT RIGHT_BUTTON
  nomouserightup:
  cmp al, MOUSE_CENTER
  jne nomousecenterup
  and byte ptr cs:mouse_button_status, NOT CENTER_BUTTON
  nomousecenterup:

  cmp al, KEY_ENABLEMOUSE
  jne nomousedisable
  and byte ptr cs:mouse_button_status, NOT MOUSE_ENABLED
;  jmp keyboard_handler_chain
  nomousedisable:
  jmp keyboard_handler_chain

  stop_hor:
  mov word ptr cs:mouse_x_speed, 0
  and byte ptr cs:mouse_button_status, NOT (DIR_RIGHT OR DIR_LEFT)
  jmp keyboard_handler_end

  stop_ver:
  mov word ptr cs:mouse_y_speed, 0
  and byte ptr cs:mouse_button_status, (NOT (DIR_DOWN OR DIR_UP)) AND 0xFF
  jmp keyboard_handler_end

  keyboard_handler_chain:
  mov ax, cs:ax_save1

  db 0xEA ; FAR JuMP
  old_keyboard_handler: dd 0x00

  keyboard_handler_end:
  mov al, 0x20
  out 0x20, al
  mov ax, cs:ax_save1
  iret

timer_handler:
  test byte ptr cs:mouse_button_status, DIR_RIGHT
  jz test_left
  call accelerate_right

  test_left:
  test byte ptr cs:mouse_button_status, DIR_LEFT
  jz test_up
  call accelerate_left

  test_up:
  test byte ptr cs:mouse_button_status, DIR_UP
  jz test_down
  call accelerate_up

  test_down:
  test byte ptr cs:mouse_button_status, DIR_DOWN
  jz timer_handler_end
  call accelerate_down

  timer_handler_end:
  db 0xEA
  old_timer_handler: dd 0x00

int33_handler:
  cmp ax, 0
  je return_driver_status
  cmp ax, 3
  je return_button_status
  cmp ax, 0x0B
  je return_motion

  jmp int33_end

  return_driver_status:
  dec ax
  mov bx, 3         ; 3 buttons
  jmp int33_end

  return_button_status:
  xor bh, bh
  mov bl, byte ptr cs:mouse_button_status
  and bl, 0x07
  mov cx, word ptr cs:cursor_col
  mov dx, word ptr cs:cursor_row
  jmp int33_end

  return_motion:
  mov cx, cs:mouse_dx
  mov cs:mouse_dx, 0
  mov dx, cs:mouse_dy
  mov cs:mouse_dy, 0

int33_end:
iret

code_start:
; replace keyboard vector
mov bx, 0x09
mov di, offset old_keyboard_handler
mov si, offset keyboard_handler
call replace_vector

; and timer vector
mov bl, 0x08
mov di, offset old_timer_handler
mov si, offset timer_handler
call replace_vector

; set mouse driver interrupt
mov bl, 0x33
xor di, di
mov si, offset int33_handler
call replace_vector

; go resident
mov ax, 0x3100
mov dx, (0xF + OFFSET code_start) ;/ 0x10
mov cl, 4
shr dx, cl
int 0x21

; bx == interrupt
; di == target pointer for saving the old vector
; si == new vector (offset only)
replace_vector:
  push ds
  xor ax, ax
  mov ds, ax

  shl bx, 1
  shl bx, 1

  les ax, [bx]

  cli

  mov word ptr [bx+2], cs
  mov word ptr [bx], si

  pop ds

  test di, di
  jz replace_vector_nosave

  mov word ptr [di+2], es
  mov word ptr [di], ax

  replace_vector_nosave:

  sti

ret 

message: db "Keyboard mouse emulator v. 1.0 (c) Sami Tikkanen 2021", 0x0D, 0x0A
db "Hold CTRL key to use mouse.", 0x0D, 0x0A
db "Move mouse cursor with arrow keys.", 0x0D, 0x0A
db "Insert == Left mouse button", 0x0D, 0x0A
db "Enter == Right mouse button", 0x0D, 0x0A
db "Keypad+ == Center mouse button", 0x0D, 0x0A
db "Going resident.$"

init_driver_:
mov ah, 0x09
mov dx, offset message
int 0x21
ret

END start


