#include <stdio.h>
#include <i86.h>
#include <dos.h>

#define serial_cb_SIZE 256

#define MAX_SERIAL_PORTS 4
#define MAX_SERIAL_STR__(x) #x
#define MAX_SERIAL_STR_(x) MAX_SERIAL_STR__(x)
#define MAX_SERIAL_STR "\nThis driver supports only " MAX_SERIAL_STR_(MAX_SERIAL_PORTS) " serial ports."

int dos_close(unsigned handle);
#pragma aux dos_close = \
  "mov ah, 0x3e" \
  "int 0x21" \
  parm [bx] value [ax];


struct serial_cb
{
  unsigned char __far *buffer;
  unsigned baud_divisor;
  volatile unsigned char write_offset;
  unsigned char read_offset;
  volatile unsigned char flow_control; 
  unsigned char error;
  unsigned char word_size;
  unsigned char stop_bits;
  unsigned char parity;
};

unsigned short __based(__segname("_CODE")) baud_divisors[] = 
{
      // baud rates:  
2304, // 0 == 50 baud
1040, // 1 == 110 baud
768,  // 2 == 150 baud
384,  // 3 == 300 baud
192,  // 4 == 600 baud
96,   // 5 == 1200 baud
48,   // 6 == 2400 baud
24,   // 7 == 4800 baud
12,   // 8 == 9600 baud
6,    // 9 == 19200 baud
3,    // 10 == 38400 baud
2,    // 11 == 57600 baud
1     // 12 == 115200 baud
};


#define XON_XOFF 0x01
#define HARDWARE 0x02
#define XOFF_RECEIVED 0x04

#define PARITY_ERROR 0x01
#define FRAMING_ERROR 0x02
#define BUFFER_FULL 0x04


#define BYTE_XOFF 0x13
#define BYTE_XON 0x11

struct serial_cb __far * __based(__segname("_CODE")) serial_cb;

unsigned short __far * __based(__segname("_CODE")) serialbase = (unsigned short __far*)0x00400000L;

unsigned char __based(__segname("_CODE")) error_code = 0;

unsigned __based(__segname("_CODE")) serial_port_count;
#define SERIAL_PORT_COUNT (((*(unsigned char __far*)0x00400011L) >> 1)&0x7)

unsigned __based(__segname("_CODE")) highest_serport_number;

unsigned current_psp(void);
#pragma aux current_psp = \
  "mov ah, 0x51" \
  "int 0x21" \
  value [bx] modify [ax];

unsigned char inbyte(unsigned port);
#pragma aux inbyte = \
  "in al, dx" \
  parm [dx] value [al];

void outbyte(unsigned port, unsigned char byte);
#pragma aux outbyte = \
  "out dx, al" \
  parm [dx] [ax];

void __far *dos_alloc(unsigned bytes);
#pragma aux dos_alloc = \
  "push cx" \
  "mov cl, 4" \
  "add bx, 0xF" \
  "shr bx, cl" \
  "pop cx" \
  "mov ah, 0x48" \
  "int 0x21" \
  "jnc end" \
  "xor ax, ax" \
  "end:" \
  "xor dx, dx" \
  "xchg ax, dx" \
  parm [bx] value [dx ax];

void __far dos_free(unsigned seg);
#pragma aux dos_free = \
  "mov ah, 0x49" \
  "mov es, bx" \
  "int 0x21" \
  parm [bx] modify [es ax];

unsigned serial_receive(unsigned devnum);
#pragma aux serial_receive = \
  "mov ah, 2" \
  "int 0x14" \
  value [ax] parm [dx];

static int serial_send(unsigned devnum, unsigned char byte)
{
  while(!(inbyte(serialbase[devnum]+5) & 0x20));

  outbyte(serialbase[devnum], byte);
  if(inbyte(serialbase[devnum]+5) & 0x02)
    return 1; // overrun
  return 0;
}


inline int is_invalid_com_port(unsigned __far *devnum)
{
  if(!*devnum || *devnum > highest_serport_number || !serialbase[--*devnum])
    return 1;

  return 0;
}

void set_serialport_settings(unsigned char port, unsigned char baud_divisor, unsigned char data_bits, unsigned char stop_bits, unsigned char parity)
{
  // enable DLAB
  outbyte(serialbase[port]+3, 0x80);
  // set baud rate
  outbyte(serialbase[port]+0, baud_divisors[baud_divisor]);
  outbyte(serialbase[port]+1, baud_divisors[baud_divisor]>>8);
  // disable DLAB & set data bits, parity and stop bits
  outbyte(serialbase[port]+3, (data_bits - 5) | ((stop_bits - 1) << 2) | parity);
  // disable FIFO
  outbyte(serialbase[port]+2, 0x00);
  // set interrupt when receiving  
  outbyte(serialbase[port]+1, 0x01);
  // set RTS/DSR, IRQ enable
  outbyte(serialbase[port]+4, 0x0B);

  serial_cb[port].baud_divisor = baud_divisor;
  serial_cb[port].word_size = data_bits;
  serial_cb[port].parity = parity;
  serial_cb[port].stop_bits = stop_bits;
  serial_cb[port].error = 0;
}


static unsigned __far com_read(unsigned, unsigned char __far*, unsigned);
#pragma aux com_read value [ax] parm [bx] [dx ax] [cx] modify exact [];
static unsigned __far com_read(unsigned devnum, unsigned char __far *buff, unsigned count)
{
  unsigned retval = 0;

  if(is_invalid_com_port(&devnum) || serial_cb[devnum].error)
  {
    error_code = 1;
    return 0;
  }

  while(serial_cb[devnum].write_offset != serial_cb[devnum].read_offset)
  {
    *buff++ = serial_cb[devnum].buffer[serial_cb[devnum].read_offset++];
    retval++;
    if(retval == count) break;
  }

  if((unsigned char)(serial_cb[devnum].write_offset - serial_cb[devnum].read_offset) < 80)
  {
    if(serial_cb[devnum].flow_control & HARDWARE)
      outbyte(serialbase[devnum]+4, 0x0B);
    if(serial_cb[devnum].flow_control & XON_XOFF)
      serial_send(devnum, BYTE_XON);
  }


  return retval;
}

void debug(unsigned char chr);
#pragma aux debug = \
  "mov ah, 0x0E" \
  "mov bx, 0x0007" \
  "int 0x10" \
  parm [al] modify [bx];

static unsigned __far com_write(unsigned devnum, unsigned char __far *buff, unsigned count);
#pragma aux com_write parm [bx] [dx ax] [cx] value [ax] modify exact [];
static unsigned __far com_write(unsigned devnum, unsigned char __far *buff, unsigned count)
{
  unsigned retval = 0;
  unsigned n;

  if(is_invalid_com_port(&devnum))
  {
    error_code = 1;
    return 0;
  }

  while(count--)
  {
    if(serial_cb[devnum].flow_control & XOFF_RECEIVED)
      break;
    if(serial_cb[devnum].flow_control & HARDWARE)
      if(!(inbyte(serialbase[devnum]+6) & 0x10)) // CTS line enabled
        break; 

    if(serial_send(devnum, *buff++))
      break;
    retval++;
  }

  return retval;
}

unsigned __far get_error(unsigned devnum);
#pragma aux get_error parm [bx] value [ax];
unsigned __far get_error(unsigned devnum)
{
  unsigned retval = error_code;
  error_code = 0;
  return retval;
}

int __far driver_command(unsigned, unsigned short __far*);
#pragma aux driver_command parm [bx] [dx ax] value [ax];
int __far driver_command(unsigned devnum, unsigned short __far *cmd)
{
  #define CMD_OPCODE cmd[0]
  #define CMD_BAUDS cmd[1]
  #define CMD_PARITY cmd[2]
  #define CMD_WORDSIZE cmd[3]
  #define CMD_STOPBITS cmd[4]
  #define CMD_FLOWCTRL cmd[5]

  if(is_invalid_com_port(&devnum))
    return 1;

  if(CMD_OPCODE)
  {
    set_serialport_settings(devnum, CMD_BAUDS, CMD_WORDSIZE, CMD_STOPBITS, CMD_PARITY);
    serial_cb[devnum].flow_control = CMD_FLOWCTRL;
  }

  CMD_BAUDS = serial_cb[devnum].baud_divisor;
  CMD_PARITY = serial_cb[devnum].parity;
  CMD_WORDSIZE = serial_cb[devnum].word_size;
  CMD_STOPBITS = serial_cb[devnum].stop_bits;
  CMD_FLOWCTRL = serial_cb[devnum].flow_control;

  return 0;
}

unsigned long __far driver_ioctl(unsigned, unsigned, unsigned char __far*, unsigned);
#pragma aux driver_ioctl parm [bx] [cx] [dx ax] [si] value [dx ax];

#define CHAR_DEVICE 0x80L
#define NO_EOF_ON_INPUT 0x40L
#define BINARY_MODE 0x20L

unsigned long __far driver_ioctl(unsigned devnum, unsigned count,
                                 unsigned char __far *buff, unsigned function)
{
  unsigned return_word;
  unsigned char eof;
  if(is_invalid_com_port(&devnum))
  {
    error_code = 0x06;
    return (long)FP_SEG(buff) << 16;
  }
  eof = (serial_cb[devnum].write_offset == serial_cb[devnum].read_offset);
  switch(function & 0xFF)
  {
    case 0: case 1:
      return_word = CHAR_DEVICE | BINARY_MODE;
      if(!eof)
        return_word |= NO_EOF_ON_INPUT;
      return ((long)return_word << 16) | function;
    case 6:
      if(eof)
        return ((long)FP_SEG(buff) << 16) | (function & 0xFF00);
      else
        return ((long)FP_SEG(buff) << 16) | (function | 0xFF);
    case 7:
      if((serial_cb[devnum].flow_control & XOFF_RECEIVED) ||
         ((serial_cb[devnum].flow_control & HARDWARE) && 
          !(inbyte(serialbase[devnum]+6) & 0x10)))
        return ((long)FP_SEG(buff) << 16) | (function & 0xFF00);
      else
        return ((long)FP_SEG(buff) << 16) | (function | 0xFF);
  }
}


int install_driver(char __far *, void __far *);
#pragma aux install_driver = \
  "mov ah, 0x6A" \
  "xor cx, cx" \
  "push ds" \
  "push es" \
  "pop ds" \
  "int 0x21" \
  "pop ds" \
  "jnc noerror" \
  "mov ax, -1" \
  "jmp end" \
  "noerror:" \
  "xor ax, ax" \
  "end:" \
  parm [dx bx] [es di] value [ax] modify [cx];

  void __far * __based(__segname("_CODE")) function_table[] =
  {
    com_read,
    com_write,
    0, // seek
    0, // open
    0, // close
    get_error, // get_error_code
    driver_command, // send_command
    driver_ioctl // ioctl
  };


#define DEFAULT_BAUD_DIVISOR 6
#define DEFAULT_DATA_BITS 8
#define DEFAULT_STOP_BITS 1

#define NO_PARITY 0
#define ODD_PARITY (1 << 3)
#define EVEN_PARITY (3 << 3)
#define MARK_PARITY (5 << 3)
#define SPACE_PARITY (7 << 3)


void serial_receiver(unsigned port)
{
  unsigned char c;

  do
  {
    if(serialbase[port])
    {
      if(!(inbyte(serialbase[port]+2) & 0x01)) // interrupt caused by this port
      {
        c = inbyte(serialbase[port]+5);
        if(c & 0x8C) // error detected
        {
          if(c & 0x04) serial_cb[port].error |= PARITY_ERROR;
          if(c & 0x08) serial_cb[port].error |= FRAMING_ERROR;
        }
        else if(c & 0x01) // data ready and the interrupt was not caused by break
        {
          if(serial_cb[port].write_offset+1 == serial_cb[port].read_offset)
          {
            serial_cb[port].error |= BUFFER_FULL;
            continue;
          }
          c = inbyte(serialbase[port]);
          serial_cb[port].buffer[serial_cb[port].write_offset++] = c;

          if(serial_cb[port].flow_control & XON_XOFF)
          {
            if(c == BYTE_XOFF)
            {
              serial_cb[port].flow_control |= XOFF_RECEIVED;
              continue;
            }
            else if(c == BYTE_XON)
            {
              serial_cb[port].flow_control &= ~XOFF_RECEIVED;
              continue;
            }
          }

          if((unsigned char)(serial_cb[port].write_offset - serial_cb[port].read_offset) == 128)
          {
            if(serial_cb[port].flow_control & XON_XOFF)
              serial_send(port, BYTE_XOFF);
            if(serial_cb[port].flow_control & HARDWARE)
              outbyte(serialbase[port]+4, 0x08); // activate -DTR and -RTS
          }
        }
      }
    }
  }
  while((port+=2) < serial_port_count);
}

void __interrupt int_ch_handler(void)
{
  serial_receiver(0);
  outbyte(0x20, 0x20);
}

void __interrupt int_bh_handler(void)
{
  serial_receiver(1);
  outbyte(0x20, 0x20);
}

int serial_initialize(void)
{
  unsigned n,i;

  serial_port_count = SERIAL_PORT_COUNT;

  printf("%u serial ports detected.\n", serial_port_count);

  if(!serial_port_count) return -2;

  for(highest_serport_number=0,n=0,i=serial_port_count;i;n++)
    if(serialbase[n])
    {
      highest_serport_number=n+1;
      i--;
    }

  printf("Highest serial port number is %u", highest_serport_number);

  if(highest_serport_number != serial_port_count)
    printf(", but there are only %u serial ports. Fix your jumper\
 configuration.", serial_port_count);

  if(highest_serport_number > MAX_SERIAL_PORTS)
  {
    puts(MAX_SERIAL_STR);
    highest_serport_number = MAX_SERIAL_PORTS;
  }


  serial_cb = dos_alloc(highest_serport_number*sizeof(struct serial_cb));
  if(!serial_cb) return -1;

  for(n=highest_serport_number;n--;)
    if(serialbase[n])
    {
      if(!(serial_cb[n].buffer = dos_alloc(serial_cb_SIZE)))
        return -1;
      serial_cb[n].write_offset = 0;
      serial_cb[n].read_offset = 0;
      serial_cb[n].flow_control = 0;

      set_serialport_settings(n, DEFAULT_BAUD_DIVISOR, DEFAULT_DATA_BITS, DEFAULT_STOP_BITS, EVEN_PARITY);
    }

  puts("\nAllocated receive buffers for serial ports.");

  _dos_setvect(0x0B, int_bh_handler);
  _dos_setvect(0x0C, int_ch_handler);

  outbyte(0x21, inbyte(0x21) & ~0x18);

  return 0;  
}

void set_memory_allocation_strategy(unsigned strategy);
#pragma aux set_memory_allocation_strategy = \
  "mov ax, 0x5801" \
  "int 0x21" \
  parm [bx];

int main()
{
  unsigned env_seg;
  unsigned char __far *psp_ptr = MK_FP(current_psp(), 0);
  int rc;
  unsigned n;

  puts("Installing COM driver.");

  set_memory_allocation_strategy(2);
  rc = serial_initialize();
  set_memory_allocation_strategy(0);

  switch(rc)
  {
    case -2: puts("No serial ports detected!"); return 1;
    case -1: puts("Out of memory!"); return 1;
  }

  if(!install_driver("COM", function_table))
  {
    puts("Done, going resident...");

    env_seg = *(unsigned __far*)&psp_ptr[0x2C];
    dos_free(env_seg);

    for(n=3;n--;)
      dos_close(n);

    _dos_keep(0, 0x10 + ((FP_OFF(serial_initialize) + 0xF) >> 4));
  }
  puts("Error");
  return 1;
}
