#include "globals.c"
#include "macros.c"
#include "structs.c"

inline struct fw_buffer __far *find_last_buffer(struct file_handle __far *handle_ptr)
{
  struct fw_buffer __far *retval;
  if(!handle_ptr->write_buffer)
    return 0;
  else
  {
    retval = MK_FP(handle_ptr->write_buffer, 0);
    while(retval->next)
      retval = MK_FP(retval->next, 0);
    return retval;
  }
}

#define MIN_BUFFER_SIZE 2048

void shrink_buffer(struct fw_buffer __far *buff_ptr, unsigned count)
{
  buff_ptr->alloc_bytes = count;
  shrink_alloc(buff_ptr->buff_seg, (buff_ptr->alloc_bytes+0xF)>>4);
}

static int add_write_buffer(unsigned handle_seg, unsigned byte_count, void __far *buff)
{
  unsigned wb_paras;
  unsigned wb_seg;
  struct file_handle __far *handle_ptr;
  struct fw_buffer __far *buff_ptr;
  struct fw_buffer __far *prev_buff;
  struct partition __far *part_ptr;
  unsigned char drive;
  long seek_bytes;

  handle_ptr = MK_FP(handle_seg, 0);

  seek_bytes = handle_ptr->offset - handle_ptr->length;

  drive = handle_ptr->file[0];

  part_ptr = MK_FP(PARTITION_SEG[drive], 0);

  if(part_ptr->buffer_bytes + seek_bytes + byte_count > free_bytes(drive))
  {
    error:
    flush_handle(handle_seg);
    return 1;
  }
  if(seek_bytes < 0) seek_bytes = 0;

  prev_buff = find_last_buffer(handle_ptr);
  if(prev_buff)
  {
    if((long)prev_buff->byte_count + byte_count <= prev_buff->alloc_bytes
       && handle_ptr->offset == prev_buff->offset + prev_buff->byte_count)
    {
      memcpy(MK_FP(prev_buff->buff_seg, prev_buff->byte_count), buff, byte_count);
      prev_buff->byte_count += byte_count;
      goto end;
    }
    else
    {
      shrink_buffer(prev_buff, prev_buff->byte_count);
    }
  }

  wb_seg = new_alloc((sizeof(struct fw_buffer)+0xF)>>4, 0);
  if(!wb_seg) goto error;

  if(byte_count <= MIN_BUFFER_SIZE)
    wb_paras = MIN_BUFFER_SIZE >> 4;
  else
  {
    wb_paras = byte_count >> 4;
    if(byte_count & 0xF) wb_paras++;
  }

  buff_ptr = MK_FP(wb_seg, 0);
  if(!(buff_ptr->buff_seg = new_alloc(wb_paras, 0)))
  {
    delete_alloc(wb_seg);
    goto error;
  }

  buff_ptr->offset = handle_ptr->offset;
  buff_ptr->byte_count = byte_count;
  buff_ptr->alloc_bytes = wb_paras << 4;
  buff_ptr->handle_seg = handle_seg;
  buff_ptr->seek_bytes = seek_bytes;
  buff_ptr->next = 0;
  part_ptr->buffer_bytes += seek_bytes;
  memcpy(MK_FP(buff_ptr->buff_seg, 0), buff, byte_count);

  // reinitialize this pointer because new_alloc() may have flushed buffers
  prev_buff = find_last_buffer(handle_ptr);

  if(!prev_buff)
  {
    handle_ptr->write_buffer = wb_seg;
    memcpy(&buff_ptr->file_fss_info, &handle_ptr->file_fss_info, sizeof(union file_fss_info));
  }
  else
  {
    prev_buff->next = wb_seg;
    memset(&buff_ptr->file_fss_info, 0, sizeof(union file_fss_info));
  }

  buffers++;

  end:
  handle_ptr->buffer_size += byte_count;
  part_ptr->buffer_bytes += byte_count;

  return 0;
}

static int flush_buffer(unsigned buff_seg)
{
  struct fw_buffer __far *fw_buff =  MK_FP(buff_seg, 0);
  struct file_handle __far *handle_ptr = MK_FP(fw_buff->handle_seg, 0);
  struct partition __far *part_ptr = MK_FP(PARTITION_SEG[handle_ptr->file[0]], 0);
  unsigned byte_count = fw_buff->byte_count;
  unsigned wr;
  unsigned long np;
  unsigned char __far *ptr;

  handle_ptr->offset = fw_buff->offset;

  memcpy(&handle_ptr->file_fss_info, &fw_buff->file_fss_info, sizeof(union file_fss_info));

  wr = write_file(fw_buff->offset, handle_ptr->file, byte_count,
     ptr = MK_FP(fw_buff->buff_seg, 0), fw_buff->handle_seg);
  np = handle_ptr->offset + wr;
  if(np > handle_ptr->length)
    handle_ptr->length = np;

  part_ptr->buffer_bytes -= wr;
  handle_ptr->buffer_size -= wr;

  if(wr != byte_count)
  {
    fw_buff->byte_count -= wr;
    fw_buff->offset += wr;
    memcpy(ptr, &ptr[wr], fw_buff->byte_count);
    shrink_buffer(fw_buff, fw_buff->byte_count);
    return 1;
  }

  part_ptr->buffer_bytes -= fw_buff->seek_bytes;

  handle_ptr->write_buffer = fw_buff->next;

  delete_alloc(fw_buff->buff_seg);  
  delete_alloc(buff_seg);

  buffers--;

  return 0;
}

unsigned handle_buffer_size(unsigned handle_seg)
{
  struct file_handle __far *handle_ptr;

  if(!handle_seg) return 0;
  handle_ptr = MK_FP(handle_seg, 0);

  return handle_ptr->buffer_size;
}

extern unsigned initial_mcb;

int flush_handle(unsigned handle_seg)
{
  struct file_handle __far *handle_ptr = MK_FP(handle_seg, 0);
  unsigned fwbuff_seg;
  unsigned next_fwbuff;
  struct fw_buffer __far *fwbuff_ptr;
  unsigned long offset;
  unsigned long next_offset = -1;
  int retval = 0;

  if(handle_ptr->write_buffer)
  {
    retval++;
    offset = handle_ptr->offset;
    fwbuff_seg = handle_ptr->write_buffer;
    do
    {
      fwbuff_ptr = MK_FP(fwbuff_seg, 0);
      next_fwbuff = fwbuff_ptr->next;
      if(next_offset == fwbuff_ptr->offset)
        memcpy(&fwbuff_ptr->file_fss_info, &handle_ptr->file_fss_info, sizeof(union file_fss_info));
      next_offset = fwbuff_ptr->offset + fwbuff_ptr->byte_count;
      if(flush_buffer(fwbuff_seg))
      {
        handle_ptr->offset = offset;
        retval = -1;
        break;
      }
    }
    while(fwbuff_seg = next_fwbuff);
    if(offset != next_offset)
      memset(&handle_ptr->file_fss_info, 0, sizeof(union file_fss_info));
    handle_ptr->write_buffer = fwbuff_seg;
    handle_ptr->offset = offset;
  }

  return retval;
}

void free_buffers(unsigned handle_seg)
{
  struct file_handle __far *handle_ptr = MK_FP(handle_seg, 0);
  struct fw_buffer __far *fwbuff_ptr;
  struct partition __far *part_ptr = MK_FP(PARTITION_SEG[handle_ptr->file[0]], 0);

  while(fwbuff_ptr = MK_FP(handle_ptr->write_buffer, 0))
  {
    handle_ptr->write_buffer = fwbuff_ptr->next;
    part_ptr->buffer_bytes -= fwbuff_ptr->seek_bytes + fwbuff_ptr->byte_count;
    delete_alloc(fwbuff_ptr->buff_seg);
    delete_alloc(FP_SEG(fwbuff_ptr));
  }
}

int flush_buffers(void)
{
  unsigned handle_seg = initial_mcb + MCB_PARAGRAPHS;
  int rc;
  unsigned retval;

  if(DONT_FLUSH_CACHES)
    return 0;

  if(retval = buffers)
  {
    while(handle_seg = next_handle_seg(handle_seg))
    {
      if((rc = flush_handle(handle_seg)) == -1)
      {
        retval = rc;
        break;
      }
    }
  }
  return retval;
}

int write_to_buffer(unsigned handle_seg, unsigned count, void __far *data)
{
  unsigned s;

  if(!FILE_BUFFER_SIZE)
  {
    return flush_handle(handle_seg) | 1;
  }

  if(cache_size <= (CACHE_SIZE>>4)
    || (s = handle_buffer_size(handle_seg) + count) < count
    || s >= FILE_BUFFER_SIZE)
  {
    if(flush_buffers() == -1)
      return -1;
    allocate_more_caches();
    if(cache_size <= (CACHE_SIZE>>4))
      return 1;
  }
  return add_write_buffer(handle_seg, count, data);
}
