.8086

.MODEL small

public init_debug
public compare_kernel_
;extern _code_end:BYTE
extern putstr_:BYTE
extern hex2str_:BYTE
extern stack_top:BYTE
extern get_vector_:BYTE
extern set_vector_:BYTE
extern _diskette_last_access:BYTE

_TEXT               SEGMENT BYTE PUBLIC USE16 'CODE'
                ASSUME CS:_TEXT                               

kernel_size proc near
  push cx
;  mov ax, offset code_end_
  mov ax, offset _diskette_last_access
  add ax, 0xF
  mov cl, 4
  shr ax, cl ; ax = size of kernel in paragraphs
  pop cx
  ret
kernel_size endp

copy_kernel proc near
  push cx
  push ax
  push bx
  push es
  push si
  push di
  push ds

  mov ax, cs
  mov ds, ax

  call kernel_size
  mov bx, 0xA000
  sub bx, ax
  mov es, bx ; es = target segment

  xor si, si
  xor di, di
;  mov cx, offset code_end_
  mov cx, offset _diskette_last_access
  mov bx, cx
  rep movsb

  pop ds
  pop di
  pop si
  pop es
  pop bx
  pop ax
  pop cx
  ret
copy_kernel endp

; make identical copy of the kernel to the end of the memory space
; (copy from CS:0)
init_debug proc near
  push ax
  push bx
  push cx
  push si
  push di

  call copy_kernel

  mov cl, 10
  add bx, 0xF
  shr bx, cl ; bx = size of kernel in kilobytes

  mov ax, 0x40
  mov es, ax
  sub es:[0x13], bx ; substract memory size in bios data table

  ; set int 1 handler
  xor ax, ax
  mov es, ax
  mov es:[6], cs
  mov es:[4], offset trap_handler

  ; set int 0x13 wrapper
  mov bx, 0x13
  call near ptr get_vector_
  mov word ptr cs:[int13_handler+2], es
  mov word ptr cs:[int13_handler], si
  push cs
  pop es
  mov si, offset int13_wrapper
  call near ptr set_vector_

  ; set int 0x12
  mov bx, 0x12
  call near ptr get_vector_
  mov word ptr cs:[int12_handler+2], es
  mov word ptr cs:[int12_handler], si
  push cs
  pop es
  mov si, offset int12_wrapper
  call near ptr set_vector_

  ; set int 0x10
  mov bx, 0x10
  call near ptr get_vector_
  mov word ptr cs:[int10_handler+2], es
  mov word ptr cs:[int10_handler], si
  push cs
  pop es
  mov si, offset int10_wrapper
  call near ptr set_vector_

  ; set int 0x16
  mov bx, 0x16
  call near ptr get_vector_
  mov word ptr cs:[int16_handler+2], es
  mov word ptr cs:[int16_handler], si
  push cs
  pop es
  mov si, offset int16_wrapper
  call near ptr set_vector_

  ; set int 0x1A
  mov bx, 0x1A
  call near ptr get_vector_
  mov word ptr cs:[int1A_handler+2], es
  mov word ptr cs:[int1A_handler], si
  push cs
  pop es
  mov si, offset int1A_wrapper
  call near ptr set_vector_

  ; toggle single step mode
  pushf
  pop ax
  or ah, 1
  push ax
  popf

  pop di
  pop si
  pop cx
  pop bx
  pop ax
  ret 
init_debug endp

; ax == 0 if kernels are equal
compare_kernel_ proc near
  push bx
  push cx
  push di
  push si
  push es
  push ds

  mov ax, cs
  mov ds, ax

  mov bx, 0xA000
  call kernel_size
  sub bx, ax

  mov es, bx ; es = segment of kernel copy
;  mov cx, offset code_end_
  mov cx, offset _diskette_last_access
  push cx

  mov di, offset stack_top
  mov si, di
  sub cx, si

  repe cmpsb
  pop ax
;  mov ax, cx
  sub ax, cx

  pop ds
  pop es
  pop si
  pop di
  pop cx
  pop bx
  ret
compare_kernel_ endp

kernel_panic: db "Kernel code changed! IP: ", 0
address_chg: db ", Change addr: ", 0
newline: db 0x0D, 0x0A, 0x00

trap_handler proc near
  cli
  push ax
  cld
  call compare_kernel_
;  test ax, 0xFFFF
;  cmp ax, offset code_end_
  cmp ax, offset _diskette_last_access
  jz no_panic

  push si
  push es
  push bp
  push bx

;  add ax, offset stack_top
  push ax ; changed address

  mov si, offset kernel_panic
  push cs
  pop es
  call near ptr putstr_

  mov bp, sp
  mov bx, ss:[bp+12]

  xchg bh, bl
  mov ax, bx
  call near ptr hex2str_
  mov si, ax
  call near ptr putstr_
  xchg bh, bl
  mov ax, bx
  call near ptr hex2str_
  mov si, ax
  call near ptr putstr_

  mov si, offset address_chg
  call near ptr putstr_
  pop bx
  inc bx
  xchg bh, bl
  mov ax, bx
  call near ptr hex2str_
  mov si, ax
  call near ptr putstr_
  xchg bh, bl
  mov ax, bx
  call near ptr hex2str_
  call near ptr putstr_

  mov si, offset newline
  call near ptr putstr_
cli
hlt
  call copy_kernel

  pop bx
  pop bp
  pop es
  pop si  

  no_panic:
  pop ax
  sti
  iret
trap_handler endp

int13_wrapper proc near
  push bp

  pushf
  mov bp, sp
  and word ptr ss:[bp+8], 0xFFFE

  db 0x9A ; call
  int13_handler: dw 0x00, 0x00

  mov bp, sp
  jnc int13_wrapper_no_carry
  or word ptr ss:[bp+6], 0x0001
  int13_wrapper_no_carry:
  or word ptr ss:[bp+6], 0x0100
  
  pop bp
  iret
int13_wrapper endp

int12_wrapper proc near
  pushf
  db 0x9A
  int12_handler: dw 0x00, 0x00
  iret
int12_wrapper endp

int1A_wrapper proc near
  pushf
  db 0x9A
  int1A_handler: dw 0x00, 0x00
  iret
int1A_wrapper endp

int10_wrapper proc near
  pushf
  db 0x9A
  int10_handler: dw 0x00, 0x00
  iret
int10_wrapper endp

int16_wrapper proc near
  push bp

  pushf
  mov bp, sp
  and word ptr ss:[bp+8], NOT 0x0040

  db 0x9A ; call
  int16_handler: dw 0x00, 0x00

  mov bp, sp
  jnc int16_wrapper_no_zero
  or word ptr ss:[bp+6], 0x0040
  int16_wrapper_no_zero:
  or word ptr ss:[bp+6], 0x0100
  
  pop bp
  iret
int16_wrapper endp


_TEXT ENDS

END
