#include "globals.c"
#include "macros.c"
#include "structs.c"

inline unsigned detect_harddisk(unsigned char drive, struct disk *retval)
{
  unsigned char sectors, heads;
  unsigned char status;

  drive |= 0x80;
 
  __asm
  {
    mov ah, 0x08
    mov dl, drive
    int 0x13
    mov sectors, cl
    and sectors, 0x3F
    mov heads, dh
    mov ah, 0x01 // some BIOSes have a bug and need this
    mov dl, drive
    int 0x13
    mov status, ah

    test ah, ah
    jnz end

#ifndef DISABLE_LBA
    // extended int13 services installation check
    mov ah, 0x41
    mov dl, drive
    mov bx, 0x55AA
    int 0x13
    jc end
    cmp bx, 0xAA55
    jne end
    test cl, 0x0001 // is device access using the packet structure supported?
    jz end // if not, jump to end
    cmp ah, 0x21 // is the enhanced disk drive spec ver at least 1.1?
    jl end
    mov sectors, 0xFF // else sectors = -1
#endif

    end:
  }

  retval->sectors = sectors;
  retval->heads = heads+1;
  
  return status;
}

char __based(__segname("_CODE")) floppydrives[] = " floppy drives";
char __based(__segname("_CODE")) harddisks[] = " hard disks";
char __based(__segname("_CODE")) detected[] = " detected\r\n";
char __based(__segname("_CODE")) mountedhd1[] = \
  "Mounted boot filesystem as partition ";
char __based(__segname("_CODE")) usinglba[] = \
  "Using LBA for hard drive ";
extern char __based(__segname("_CODE")) newline[];

int initdisk(void)
{
  unsigned n, h;
  unsigned char status;
  struct disk disk_geometry;
  unsigned char bios_equipment_flags = *(unsigned char __far*)MK_FP(0x40, 0x10);
  unsigned char bios_floppy_count = bios_equipment_flags & 0x01;

  RAWCACHE_LAST_DISK = -1;

  if(bios_floppy_count) bios_floppy_count += bios_equipment_flags >> 6;

  FLOPPY_COUNT = bios_floppy_count;

  DRIVE_TABLE = new_alloc((MAX_PARTITIONS * DRIVE_STRUCT_SIZE + 0xF)>>4, 0);
  if(!DRIVE_TABLE) return -1;

  set_mcb_type(DRIVE_TABLE, MCB_SYSTEM);
  memset(DRIVE_TABLE_ARR, 0, MAX_PARTITIONS*DRIVE_STRUCT_SIZE);

  for(n=0;n<bios_floppy_count;n++)
  {
    DRIVE_TABLE_ARR[n*DRIVE_STRUCT_SIZE] |= DRIVE_REMOVABLE_BIT | DRIVE_FLOPPY_BIT;
    add_partition(n, 0, -1, 0, MAX_SECTOR_SIZE, 0);
  }

  putstr(hex2str(FLOPPY_COUNT));
  putstr(floppydrives);
  putstr(detected);
  putstr(hex2str(HARD_DISK_COUNT));
  putstr(harddisks);
  putstr(detected);

  set_system_drive();
    
  // detect hard disks
  if(HARD_DISK_COUNT)
  {
    for(n=0;n<HARD_DISK_COUNT;n++)
    {
      status = detect_harddisk(n, &disk_geometry);
      if(!status)
      {
        h = FLOPPY_COUNT_R + n;
        HD_DYNPART_COUNT++;
        SAVE_DRIVE_SECTOR_SIZE(h, HD_SECTOR_SIZE);
        if(0xFF == (DRIVE_SECTORS(h) = disk_geometry.sectors))
        {
          putstr(usinglba);
          putstr(hex2str(n));
          putstr(newline);
        }
        DRIVE_HEADS(h) = disk_geometry.heads;
        if(h == SYSTEM_DRIVE)
        { // if the boot drive is a hard disk, mount the boot filesystem
          h = add_partition(h, BOOT_PARTITION_OFFSET, -1, 0, HD_SECTOR_SIZE, 0);
          putstr(mountedhd1);
          putch('A' + h);
          putstr(newline);
        }
      }
    }
  }

  return 0;
}


