#include "globals.c"
#include "macros.c"
#include "structs.c"
// 13-1 words
#define EXE_HEADER_SIZE 12

unsigned move_program_image(unsigned psp, unsigned req_paras, unsigned program_paras)
{
  unsigned ret_seg = new_alloc(req_paras, 0);
  unsigned char __far *old_code_ptr = MK_FP(psp, 0);
  unsigned char __far *new_code_ptr = MK_FP(ret_seg, 0);
  unsigned long copy_count = (unsigned long)program_paras << 4;

  if(ret_seg)
  {
    while(1)
      if(copy_count >= 0xFFF0)
      {
        memcpy(new_code_ptr, old_code_ptr, 0xFFF0);
        copy_count -= 0xFFF0;
        new_code_ptr += 0xFFF0;
        old_code_ptr += 0xFFF0;
        fix_farptr_offset((void __far**)&new_code_ptr);
        fix_farptr_offset((void __far**)&old_code_ptr);
      }
      else
      {
        memcpy(new_code_ptr, old_code_ptr, copy_count);
        break;
      }
  }

  delete_alloc(psp);

  return ret_seg;
}

static void replace_relocations(unsigned reloc_seg, unsigned reloc_cnt,
                                unsigned reloc_add, unsigned image_seg)
{
  unsigned __far *reloc_ptr;
  unsigned long __far *relocations = MK_FP(reloc_seg, 0);
  unsigned n;

  for(n=reloc_cnt;n--;)
  {
    reloc_ptr = (void __far*)&relocations[n];
    *(unsigned __far*)MK_FP(reloc_ptr[1]+image_seg, reloc_ptr[0])
      += reloc_add;
  }
}

unsigned load_exe(dirindex_t *program, struct mzload *mz_struct)
{
  unsigned header[EXE_HEADER_SIZE];
  unsigned long offset;
  unsigned n;
  unsigned long file_size;
  unsigned program_paras;
  unsigned ret_seg = 0;
  unsigned read_count;
  unsigned char __far *code_ptr;
  unsigned relocation_seg;
  unsigned relocation_paras;
  unsigned max_paras;
  unsigned image_start;
#define relocations MK_FP(relocation_seg, 0)
#define last_page_size header[0]
#define pages header[1]
#define relocation_count header[2]
#define header_size header[3]
#define min_paras header[4]
#define req_paras header[5]
#define hdinitial_ss header[6]
#define hdinitial_sp header[7]
#define file_checksum header[8]
#define hdinitial_ip header[9]
#define hdinitial_cs header[10]
#define relocation_offset header[11]
  
  if(read_file(2, program, (EXE_HEADER_SIZE<<1), (void __far*)header, 0)
    != (EXE_HEADER_SIZE<<1))
  {
    error_code = 11;
    return 0;
  }
  

  file_size = (((unsigned long)pages-1) << 9) + last_page_size;
  max_paras = free_paras(1, 1, 0);
  if(!mz_struct->load_overlay)
  {
    program_paras = file_size >> 4;
    program_paras -= header_size; // size of program image
    program_paras += 0x10; // add psp
    if(file_size & 0xF) program_paras++;

    req_paras += program_paras;
    min_paras += program_paras;
    if(req_paras < program_paras) req_paras = -1;

    if(max_paras >= req_paras)
      mz_struct->paras = req_paras;
    else if(max_paras >= min_paras)
      mz_struct->paras = max_paras;
    else
    {
      mz_struct->paras = min_paras;
      nomemory:
      error_code = 0x08;
      return 0;
    }
  }
  else
    program_paras = 0;

  relocation_paras = ((relocation_count<<2)+0xF)>>4;
  if(max_paras - program_paras < relocation_paras)
    goto nomemory;

  ret_seg = new_alloc(program_paras + relocation_paras, 1);
  relocation_seg = ret_seg + program_paras;

  // load relocations
  if(read_file(relocation_offset, program, relocation_count<<2, relocations, 0)
     != relocation_count<<2)
  {
    goto error2;
  }

  offset = header_size << 4;
  
  // load program image
  if(!mz_struct->load_overlay)
    image_start = ret_seg + 0x10;
  else
    image_start = mz_struct->initial_cs;

  code_ptr = MK_FP(image_start, 0);  
  while(1)
  {
    if((file_size - offset) >> 16)
      read_count = 0xFFFF;
    else
      read_count = file_size - offset;

    n = read_file(offset, program, read_count, code_ptr, 0);

    if(n != read_count) break;
    offset += n;
    if(offset == file_size) break;
    add_charptr(&code_ptr, n);
  }
  if(offset != file_size) // too small file
  {
    error2:
    if(!error_code) error_code = 0x0D;
    delete_alloc(ret_seg);
    return 0;
  }

  if(!mz_struct->load_overlay)
  {
    if(modify_alloc(ret_seg, mz_struct->paras) != mz_struct->paras)
      if(!(ret_seg = move_program_image(ret_seg, mz_struct->paras, program_paras + relocation_paras)))
        return 0;
      else
      {
        image_start = ret_seg + 0x10;
        relocation_seg = ret_seg + program_paras;
      }
    replace_relocations(relocation_seg, relocation_count, image_start, image_start);

    mz_struct->initial_ss = hdinitial_ss + image_start;
    mz_struct->initial_cs = hdinitial_cs + image_start;
    mz_struct->initial_sp = hdinitial_sp;
    mz_struct->initial_ip = hdinitial_ip;
  }
  else
  {
    replace_relocations(relocation_seg, relocation_count, mz_struct->initial_ip, image_start);
    delete_alloc(ret_seg);
  }

  return ret_seg;
}
