#include "globals.c"
#include "macros.c"
#include "structs.c"

dirindex_t path_last_index(dirindex_t __far *path)
{
  unsigned n = 0;
  while(path[n+1] != -1)
    n++;
  return path[n];
}

void handle_file(unsigned handle, dirindex_t index, unsigned force)
{
  unsigned handle_seg = find_handle(CURRENT_PSP, handle);
  struct file_handle __far *handle_ptr = MK_FP(handle_seg, 0);
  unsigned n;
  struct file_info fileinfo;
  unsigned char drive = WORKING_DRIVE;

/*  if(!handle_seg) // this should never fail...
  {
    error_code = 0x06;
    return;
  }*/

  handle_ptr->file[0] = drive;
  n = getcwd(drive, &handle_ptr->file[1]);
  handle_ptr->file[n++] = index;
  handle_ptr->file[n] = -1;

  fileinfo = file_info(drive, index);
  if((fileinfo.attr & ATTR_DIR) || 
     (!force && (fileinfo.attr & ATTR_RDONLY) && (handle_ptr->type & 0x02)))
  {
    delete_handle(CURRENT_PSP, handle, 0);
    error_code = 0x05;
    return;
  }
  handle_ptr->time = fileinfo.time;
  handle_ptr->date = fileinfo.date;

  handle_ptr->length = fileinfo.size;
  handle_ptr->static_fss_info = get_handle_static_info(drive, index);
}

dirindex_t get_file_index(unsigned char drive, char __far *filename, dirindex_t *old_dir)
{
  int rc;
  dirindex_t index = 0;
  char entryname[13];
  unsigned char attr;
  unsigned type;

  while((rc = directory_entry(drive, index, entryname, DIRINDEX_READ_ENTRIES)) <= 0)
  {
    if(rc < -1) break;
    if(!strncmp(entryname, filename, sizeof(entryname)-1))
      break;
    index++;
  }
  if(rc == -2) // try case-insensitive search
  {
    index = 0;
    while((rc = directory_entry(drive, index, entryname, DIRINDEX_READ_ENTRIES)) <= 0)
    {
      if(rc < -1) break;
      if(!cistrncmp(entryname, filename, sizeof(entryname)-1))
        break;
      index++;
    }
  }
  if(rc == -2) error_code = 2;
  else if(rc == -3) error_code = 0x0F;
  else
  {
    error_code = 0;
    if(old_dir)
    {
      attr = file_info(drive, index).attr;
      if((attr & ATTR_SYMLINK) && !(attr & ATTR_DIR))
      {
        type = 0;
        rc = get_symlink_path(index, &index, attr, &type, old_dir);
        if(type)
        {
          // if type == -1, the symlink points to a partition
          // else type points to the driver segment
          // index is the device number
          error_code = ERR_IS_SYMLINK;
          return (index << 16) | type;
        }
      }
    }
  }

  return index;
}

// special case to find also filenames without the dot
char __based (__segname("_CODE")) allfiles[] = "*.*";

struct file_info find_file_with_wildcards(struct dta __far *dta_ptr)
{
  int rc;
  char __far * const keyword = dta_ptr->filename;
  const unsigned attrib = dta_ptr->attr;
 //  unsigned __far *ptr_to_dta = (unsigned __far*)DTA_BUFFER;
//  struct dta __far *dta_ptr = MK_FP(*ptr_to_dta, 0);
  struct file_info retval;
  unsigned char drive = dta_ptr->drive;

  while((rc = directory_entry(drive, dta_ptr->index, &DTA_BUFFER[0x1E], DIRINDEX_READ_ENTRIES)) <= 0)
  {
    if(rc == -1) goto next;
    if(rc < -1) break;
    if(!memcmp(keyword, allfiles, 4) || !wccistrcmp(keyword, &DTA_BUFFER[0x1E]))
    {
      retval = file_info(drive, dta_ptr->index);
      if(!(retval.attr & 0x1E) || (attrib & retval.attr))
        break;
    }
    next:dta_ptr->index++;
  }
  if(rc == -2) error_code = 0x12;
  else if(rc == -3) error_code = 0x0F;
  else error_code = 0;
  return retval;
}

// chdir to the directory where the filename is if the filename is a full path
// returns zero if chdir failed, otherwise returns pointer to the actual filename
char __far *change_file_dir(char __far *filename, dirindex_t *old_cwd)
{
  char __far *ptr;
  char __far *dir;
  char fnfc;

  if(ptr = strchr(filename, ':'))
    filename = ptr + 1;

  dir = filename;

  while(ptr = find_path_separator(filename))
  {
    filename = ptr+1;
  }
  if(filename != dir)
  {
    fnfc = *filename;
    *filename = 0;
    set_cwd(dir, old_cwd);
    *filename = fnfc;
  }
  if(error_code) return 0;
  return filename;
}

unsigned char drive_from_path(char __far *path)
{
  unsigned char drive;
  if(path[1] == ':')
  {
    drive = (path[0]-1) & 0x1F;
    return drive;
  }
  else return WORKING_DRIVE;
}

// if the path has a drive specifier in it, change to that drive
// returns the new working drive of -1 if the drive does not exist
int change_file_drive(char __far *filename)
{
  unsigned char drive = drive_from_path(filename);

  if(partition_exists(drive))
    chdrive(drive);
  else
    return -1;

  return WORKING_DRIVE;
}

/*
  Changes back to the old dir and drive.
*/
void chg2olddir(dirindex_t *old_cwd)
{
  chdir(WORKING_DRIVE, &old_cwd[1]);
  WORKING_DRIVE = old_cwd[0];
}

/*
  Tries to change current working dir and drive to where the file is.
  Returns 0 if an error happened. Normally returns pointer to the actual
  filename.
*/
char __far *chg2filedir(char __far *filename, dirindex_t *old_cwd)
{
  old_cwd[0] = WORKING_DRIVE; // save OLD WORKING DRIVE to old_cwd[0]
  if(change_file_drive(filename) == -1)
    goto error;
  // save OLD WORKING DIRECTORY of the NEW WORKING DRIVE to &old_cwd[1]
  getcwd(WORKING_DRIVE, &old_cwd[1]);
  filename = change_file_dir(filename, old_cwd);

  if(!filename || !*filename)
  {
    chg2olddir(old_cwd);
    error:
    error_code = 0x03;
    return 0;
  }
  return filename;
}

