#include "globals.c"
#include "macros.c"
#include "structs.c"

#define FS_CHDIR_FUNC 0
#define FS_GETCWD_FUNC 1
#define FS_FSCHGD_FUNC 2
#define FS_FREESPACE_FUNC 3
#define FS_DIRENTRY_FUNC 4
#define FS_FILEINFO_FUNC 5
#define FS_HANDLESTATINFO_FUNC 6
#define FS_READFILE_FUNC 7
#define FS_RENAME_FUNC 8
#define FS_WRITEFILE_FUNC 9
#define FS_TRUNCFILE_FUNC 10
#define FS_CREATFILE_FUNC 11
#define FS_CREATEDIR_FUNC 12
#define FS_UPDATEFILEINFO_NOHDL_FUNC 13
#define FS_UPDATEFILEINFO_HDL_FUNC 14
#define FS_SETFILEINFO_FUNC 15
#define FS_DELETE_FUNC 16
#define FS_MOUNT 17
#define FS_UNMOUNT 18
#define FS_RAWREAD 19
#define FS_RAWWRITE 20
#define FS_GETERROR 21

unsigned partition_drvseg(unsigned char drive)
{
  struct partition __far *part_ptr = MK_FP(PARTITION_SEG[drive], 0);

  if(!(part_ptr->type & 0x80))
    return 0;
  else
    return part_ptr->d.driver_seg;
}

unsigned call_fsdriver_geterror(void __far * __far *func);
#pragma aux call_fsdriver_geterror = \
  "call far ptr es:[di]" \
  parm [es di] value [ax];

static void fsdriver_reterror(unsigned drvseg, unsigned old_flags, unsigned old_psp)
{
  struct driver __far * const driver_ptr = MK_FP(drvseg, 0);
  error_code = call_fsdriver_geterror(&driver_ptr->func_table[FS_GETERROR]);
  driver_func_ret(old_flags, old_psp);
}

int call_fsdriver_func_word_lptr(void __far * __far *func, unsigned drive,
                                 void __far *);
#pragma aux call_fsdriver_func_word_lptr = \
  "call far ptr es:[di]" \
  value [ax] parm [es di] [bx] [dx ax];

unsigned long call_fsdriver_func_word_dword(void __far* __far*, unsigned, unsigned long);
#pragma aux call_fsdriver_func_word_dword = \
  "call far ptr es:[di]" \
  value [dx ax] parm [es di] [bx] [dx ax];

int call_fsdriver_func_dword(void __far* __far*, unsigned long);
#pragma aux call_fsdriver_func_dword = \
  "call far ptr es:[di]" \
  value [ax] parm [dx ax];

int call_fsdriver_func_word_nptr(void __far* __far*, unsigned, void *);
#pragma aux call_fsdriver_func_word_nptr = \
  "call far ptr es:[di]" \
  value [ax] parm [es di] [bx] [ax];

int call_fsdriver_func_word(void __far * __far *func, unsigned);
#pragma aux call_fsdriver_func_word = \
  "call far ptr es:[di]" \
  value [ax] parm [es di] [bx];

void stfs_chdir(unsigned char, dirindex_t __far*);
void chdir(unsigned char drive, dirindex_t __far *index)
{
  const unsigned drvseg = partition_drvseg(drive);
  unsigned old_indos_flag, old_psp;
  struct driver __far * driver_ptr;
  if(!drvseg)
    stfs_chdir(drive, index);
  else
  {
    driver_ptr = MK_FP(drvseg, 0);
    driver_func_entry(&old_indos_flag, &old_psp, drvseg);
    call_fsdriver_func_word_lptr(&driver_ptr->func_table[FS_CHDIR_FUNC],
                                 drive, index);
    driver_func_ret(old_indos_flag, old_psp);
  }
}

int stfs_getcwd(unsigned char, dirindex_t __far*);
int getcwd(unsigned char drive, dirindex_t __far *index)
{
  const unsigned drvseg = partition_drvseg(drive);
  unsigned old_indos_flag, old_psp;
  struct driver __far * driver_ptr;
  int rc;
  if(!drvseg)
    return stfs_getcwd(drive, index);
  else
  {
    driver_ptr = MK_FP(drvseg, 0);
    driver_func_entry(&old_indos_flag, &old_psp, drvseg);
    rc = call_fsdriver_func_word_lptr(&driver_ptr->func_table[FS_GETCWD_FUNC],
                                 drive, index);
    driver_func_ret(old_indos_flag, old_psp);
    return rc;
  }
}

int stfs_fs_changed(unsigned char);
int fs_changed(unsigned char drive)
{
  const unsigned drvseg = partition_drvseg(drive);
  unsigned old_indos_flag, old_psp;
  struct driver __far * driver_ptr;
  int rc;
  if(!drvseg)
    return stfs_fs_changed(drive);
  else
  {
    driver_ptr = MK_FP(drvseg, 0);
    driver_func_entry(&old_indos_flag, &old_psp, drvseg);
    rc = call_fsdriver_func_word(&driver_ptr->func_table[FS_FSCHGD_FUNC], drive);
    fsdriver_reterror(drvseg, old_indos_flag, old_psp);
    return rc;
  }
}

unsigned long free_bytes(unsigned char drive)
{
  struct diskfree diskfree;
  struct partition __far *part_ptr = MK_FP(PARTITION_SEG[drive], 0);

  if(part_ptr->size_known)
    return part_ptr->free_bytes;

  free_space(drive, &diskfree);
  return part_ptr->free_bytes;
}

void stfs_free_space(unsigned char, struct diskfree *);
void free_space(unsigned char drive, struct diskfree *retval)
{
  const unsigned drvseg = partition_drvseg(drive);
  unsigned old_indos_flag, old_psp;
  struct driver __far * driver_ptr;
  if(!drvseg)
    stfs_free_space(drive, retval);
  else
  {
    driver_ptr = MK_FP(drvseg, 0);
    driver_func_entry(&old_indos_flag, &old_psp, drvseg);
    call_fsdriver_func_word_nptr(&driver_ptr->func_table[FS_FREESPACE_FUNC],
                                 drive, retval);
    fsdriver_reterror(drvseg, old_indos_flag, old_psp);
  }
}

int call_fsdriver_func_dword_dword_word(void __far * __far *,
                              unsigned long, unsigned long, unsigned);
#pragma aux call_fsdriver_func_dword_dword_word = \
  "call far ptr es:[di]" \
  value [ax] parm [es di] [dx ax] [cx si] [bx];

int call_fsdriver_func_dword_fptr(void __far * __far *,
                              unsigned long, void __far*);
#pragma aux call_fsdriver_func_dword_fptr = \
  "call far ptr es:[di]" \
  value [ax] parm [es di] [bx cx] [dx ax];

int call_fsdriver_func_dword_word_nearptr(void __far * __far *,
                                       unsigned long, unsigned, void *);
#pragma aux call_fsdriver_func_dword_word_nearptr = \
  "call far ptr es:[di]" \
  value [ax] parm [es di] [dx ax] [bx] [cx];

int call_fsdriver_func_dword_word_word(void __far * __far *,
                                       unsigned long, unsigned, unsigned);
#pragma aux call_fsdriver_func_dword_word_word = \
  "call far ptr es:[di]" \
  value [ax] parm [es di] [dx ax] [bx] [cx];


int stfs_directory_entry(unsigned char, dirindex_t, unsigned char __far*, unsigned);
int directory_entry(unsigned char drive, dirindex_t index,
                    unsigned char __far *buff, unsigned count)
{
  const unsigned drvseg = partition_drvseg(drive);
  struct driver __far * driver_ptr;
  unsigned old_indos_flag, old_psp;
  int rc;
  if(!drvseg)
    return stfs_directory_entry(drive, index, buff, count);
  else
  {
    driver_ptr = MK_FP(drvseg, 0);
    driver_func_entry(&old_indos_flag, &old_psp, drvseg);
    rc = call_fsdriver_func_dword_word_nearptr(&driver_ptr->func_table[FS_DIRENTRY_FUNC],
                                        index, (count << 8) | drive, &buff);
    fsdriver_reterror(drvseg, old_indos_flag, old_psp);
    return rc;
  }
}

int call_fsdriver_func_word_nptr_dword(void __far * __far *, unsigned, void*, unsigned long);
#pragma aux call_fsdriver_func_word_nptr_dword = \
  "call far ptr es:[di]" \
  parm [es di] [bx] [cx] [dx ax] value [ax];

struct file_info stfs_file_info(unsigned char, dirindex_t);
struct file_info file_info(unsigned char drive, dirindex_t file_index)
{
  const unsigned drvseg = partition_drvseg(drive);
  struct driver __far * driver_ptr;
  unsigned old_indos_flag, old_psp;
  struct file_info retval;
  if(!drvseg)
    return stfs_file_info(drive, file_index);
  else
  {
    driver_ptr = MK_FP(drvseg, 0);
    driver_func_entry(&old_indos_flag, &old_psp, drvseg);
    call_fsdriver_func_word_nptr_dword(&driver_ptr->func_table[FS_FILEINFO_FUNC],
                                    drive, &retval, file_index);
    fsdriver_reterror(drvseg, old_indos_flag, old_psp);
    return retval;
  }
}

union stat_fss_info stfs_get_handle_static_info(unsigned char, dirindex_t);
union stat_fss_info get_handle_static_info(unsigned char drive, dirindex_t index)
{
  const unsigned drvseg = partition_drvseg(drive);
  struct driver __far * driver_ptr;
  unsigned old_indos_flag, old_psp;
  union stat_fss_info retval;
  if(!drvseg)
    return stfs_get_handle_static_info(drive, index);
  else
  {
    driver_ptr = MK_FP(drvseg, 0);
    driver_func_entry(&old_indos_flag, &old_psp, drvseg);
    retval.unused = call_fsdriver_func_word_dword(&driver_ptr->func_table[FS_HANDLESTATINFO_FUNC],
                                     drive, index);
    fsdriver_reterror(drvseg, old_indos_flag, old_psp);
    return retval;
  }
}

unsigned call_fsdriver_rw(void __far * __far *, unsigned long*, dirindex_t __far**,
                          unsigned count, void __far **buff, unsigned handle_seg);
#pragma aux call_fsdriver_rw = \
  "call far ptr es:[di]" \
  value [ax] parm [es di] [ax] [bx] [cx] [dx] [si];

unsigned stfs_read_file(unsigned long, dirindex_t __far *, unsigned count, void __far *, unsigned);
unsigned read_file(unsigned long offset, dirindex_t __far *file_index,
                   unsigned count, void __far *buff, unsigned handle_seg)
{
  const unsigned drvseg = partition_drvseg(file_index[0]);
  struct driver __far * driver_ptr;
  unsigned old_indos_flag, old_psp;
  unsigned retval;
  if(!drvseg)
    return stfs_read_file(offset, file_index, count, buff, handle_seg);
  else
  {
    driver_ptr = MK_FP(drvseg, 0);
    driver_func_entry(&old_indos_flag, &old_psp, drvseg);
    retval = call_fsdriver_rw(&driver_ptr->func_table[FS_READFILE_FUNC], &offset,
                              &file_index, count, &buff, handle_seg);
    fsdriver_reterror(drvseg, old_indos_flag, old_psp);
    return retval;
  }
}

int stfs_rename_file(unsigned char, dirindex_t, char __far*);
int rename_file(unsigned char drive, dirindex_t index, char __far *newname)
{
  const unsigned drvseg = partition_drvseg(drive);
  struct driver __far *driver_ptr;
  unsigned old_indos_flag, old_psp;
  int retval;
  if(!drvseg)
    return stfs_rename_file(drive, index, newname);
  else
  {
    driver_ptr = MK_FP(drvseg, 0);
    if(!driver_ptr->func_table[FS_RENAME_FUNC])
    {
      error_code = 1;
      return -1;
    }
    else
    {
      driver_func_entry(&old_indos_flag, &old_psp, drvseg);
      retval = call_fsdriver_func_dword_dword_word(&driver_ptr->func_table[FS_RENAME_FUNC],
                                                   index, (unsigned long)newname, drive);
      fsdriver_reterror(drvseg, old_indos_flag, old_psp);
      return retval;
    }
  }
}

unsigned stfs_write_file(unsigned long, dirindex_t __far*, unsigned count, void __far *, unsigned);
unsigned write_file(unsigned long offset, dirindex_t __far *file_index,
                    unsigned count, void __far *buff, unsigned handle_seg)
{
  const unsigned drvseg = partition_drvseg(file_index[0]);
  struct driver __far *driver_ptr;
  unsigned old_indos_flag, old_psp;
  int retval;

  if(!drvseg)
    return stfs_write_file(offset, file_index, count, buff, handle_seg);
  else
  {
    driver_ptr = MK_FP(drvseg, 0);
    if(!driver_ptr->func_table[FS_WRITEFILE_FUNC])
    {
      error_code = 1;
      return 0;
    }
    else
    {
      driver_func_entry(&old_indos_flag, &old_psp, drvseg);
      retval = call_fsdriver_rw(&driver_ptr->func_table[FS_WRITEFILE_FUNC], &offset,
                              &file_index, count, &buff, handle_seg);
      fsdriver_reterror(drvseg, old_indos_flag, old_psp);
      return retval;
    }
  }
}

int stfs_truncate_file(unsigned);
int truncate_file(unsigned handle_seg)
{
  struct file_handle __far * const handle_ptr = MK_FP(handle_seg, 0);
  const unsigned drvseg = partition_drvseg(handle_ptr->file[0]);
  struct driver __far *driver_ptr;
  unsigned old_indos_flag, old_psp;
  int rc;

  if(!drvseg)
    return stfs_truncate_file(handle_seg);
  else
  {
    driver_ptr = MK_FP(drvseg, 0);
    if(!driver_ptr->func_table[FS_TRUNCFILE_FUNC])
    {
      error_code = 1;
      return 1;
    }
    else
    {
      driver_func_entry(&old_indos_flag, &old_psp, drvseg);
      rc = call_fsdriver_func_word(&driver_ptr->func_table[FS_TRUNCFILE_FUNC], handle_seg);
      fsdriver_reterror(drvseg, old_indos_flag, old_psp);
      return rc;
    }
  }
}

dirindex_t call_fsdriver_func_word_fptr_word(void __far * __far *, unsigned, void __far*, unsigned);
#pragma aux call_fsdriver_func_word_fptr_word = \
  "call far ptr es:[di]" \
  parm [es di] [bx] [dx ax] [cx] value [dx ax];

dirindex_t stfs_create_file(unsigned char, char __far *, unsigned char);
dirindex_t create_file(unsigned char drive, char __far *name, 
                       unsigned char attr)
{
  const unsigned drvseg = partition_drvseg(drive);
  struct driver __far *driver_ptr;
  unsigned old_indos_flag, old_psp;
  int rc;

  if(!drvseg)
    return stfs_create_file(drive, name, attr);
  else
  {
    driver_ptr = MK_FP(drvseg, 0);
    if(!driver_ptr->func_table[FS_CREATFILE_FUNC])
    {
      error_code = 1;
      return -1;
    }
    else
    {
      driver_func_entry(&old_indos_flag, &old_psp, drvseg);
      rc = call_fsdriver_func_word_fptr_word(&driver_ptr->func_table[FS_CREATFILE_FUNC],
                                             drive, name, attr);
      fsdriver_reterror(drvseg, old_indos_flag, old_psp);
      return rc;
    }
  }
}

dirindex_t stfs_create_dir(unsigned char, char __far *);
dirindex_t create_dir(unsigned char drive, char __far *name)
{
  const unsigned drvseg = partition_drvseg(drive);
  struct driver __far * driver_ptr;
  unsigned old_indos_flag, old_psp;
  int rc;

  if(!drvseg)
    return stfs_create_dir(drive, name);
  else
  {
    driver_ptr = MK_FP(drvseg, 0);
    if(!driver_ptr->func_table[FS_CREATFILE_FUNC])
    {
      error_code = 1;
      return -1;
    }
    else
    {
      driver_func_entry(&old_indos_flag, &old_psp, drvseg);
      rc = call_fsdriver_func_word_dword(&driver_ptr->func_table[FS_CREATEDIR_FUNC],
                                              drive, (unsigned long)name);
      fsdriver_reterror(drvseg, old_indos_flag, old_psp);
      return rc;
    }
  }
}

int stfs_update_file_info_nohandle(dirindex_t __far *);
int update_file_info_nohandle(dirindex_t __far *file)
{
  const unsigned drvseg = partition_drvseg(file[0]);
  struct driver __far * driver_ptr;
  unsigned old_indos_flag, old_psp;
  int rc;

  if(!drvseg)
    return stfs_update_file_info_nohandle(file);
  else
  {
    driver_ptr = MK_FP(drvseg, 0);
    if(!driver_ptr->func_table[FS_CREATFILE_FUNC])
    {
      error_code = 1;
      return -1;
    }
    else
    {
      driver_func_entry(&old_indos_flag, &old_psp, drvseg);
      rc = call_fsdriver_func_dword(&driver_ptr->func_table[FS_UPDATEFILEINFO_NOHDL_FUNC],
                                              (unsigned long)file);
      fsdriver_reterror(drvseg, old_indos_flag, old_psp);
      return rc;
    }
  }
}

int stfs_update_file_info(unsigned);
int update_file_info(unsigned handle_seg)
{
  struct file_handle __far * const handle_ptr = MK_FP(handle_seg, 0);
  const unsigned drvseg = partition_drvseg(handle_ptr->file[0]);
  struct driver __far * driver_ptr;
  unsigned old_indos_flag, old_psp;
  int rc;

  if(!drvseg)
    return stfs_update_file_info(handle_seg);
  else
  {
    driver_ptr = MK_FP(drvseg, 0);
    if(!driver_ptr->func_table[FS_CREATFILE_FUNC])
    {
      error_code = 1;
      return -1;
    }
    else
    {
      driver_func_entry(&old_indos_flag, &old_psp, drvseg);
      rc = call_fsdriver_func_word(&driver_ptr->func_table[FS_UPDATEFILEINFO_HDL_FUNC],
                                              handle_seg);
      fsdriver_reterror(drvseg, old_indos_flag, old_psp);
      return rc;
    }
  }
}

int stfs_set_file_info(unsigned char, dirindex_t,  struct file_info __far*);
int set_file_info(unsigned char drive, dirindex_t index, 
                  struct file_info __far *info)
{
  const unsigned drvseg = partition_drvseg(drive);
  struct driver __far * driver_ptr;
  unsigned old_indos_flag, old_psp;
  int rc;

  if(!drvseg)
    return stfs_set_file_info(drive, index, info);
  else
  {
    driver_ptr = MK_FP(drvseg, 0);
    if(!driver_ptr->func_table[FS_CREATFILE_FUNC])
    {
      error_code = 1;
      return -1;
    }
    else
    {
      driver_func_entry(&old_indos_flag, &old_psp, drvseg);
      rc = call_fsdriver_func_dword_fptr(&driver_ptr->func_table[FS_UPDATEFILEINFO_HDL_FUNC],
                                              index, info);
      fsdriver_reterror(drvseg, old_indos_flag, old_psp);
      return rc;
    }
  }
}

int stfs_delete_file(unsigned char, dirindex_t, unsigned char);
int delete_file(unsigned char drive, dirindex_t index, unsigned char force)
{
  const unsigned drvseg = partition_drvseg(drive);
  struct driver __far * driver_ptr;
  unsigned old_indos_flag, old_psp;
  int rc;

  if(!drvseg)
    return stfs_delete_file(drive, index, force);
  else
  {
    driver_ptr = MK_FP(drvseg, 0);
    if(!driver_ptr->func_table[FS_CREATFILE_FUNC])
    {
      error_code = 1;
      return -1;
    }
    else
    {
      driver_func_entry(&old_indos_flag, &old_psp, drvseg);
      rc = call_fsdriver_func_dword_word_word(&driver_ptr->func_table[FS_UPDATEFILEINFO_HDL_FUNC],
                                              index, drive, force);
      fsdriver_reterror(drvseg, old_indos_flag, old_psp);
      return rc;
    }
  }
}

// the function must return 1 if the partition is not on a local harddisk,
// and 0 if the raw partition data can be read from file DRV:n
int fsdriver_mount_partition(unsigned driver_seg, unsigned char drive, char __far *imagefile)
{
  unsigned old_indos_flag, old_psp;
  struct driver __far * const driver_ptr = MK_FP(driver_seg, 0);
  int rc;
  driver_func_entry(&old_indos_flag, &old_psp, driver_seg);
  rc = call_fsdriver_func_word_lptr(&driver_ptr->func_table[FS_MOUNT], drive, imagefile);
  fsdriver_reterror(driver_seg, old_indos_flag, old_psp);
  return rc;
}

int fsdriver_unmount_partition(unsigned driver_seg, unsigned char drive)
{
  unsigned old_indos_flag, old_psp;
  struct driver __far * const driver_ptr = MK_FP(driver_seg, 0);
  int rc;
  driver_func_entry(&old_indos_flag, &old_psp, driver_seg);
  rc = call_fsdriver_func_word(&driver_ptr->func_table[FS_UNMOUNT], drive);
  fsdriver_reterror(driver_seg, old_indos_flag, old_psp);
  return rc;
}

unsigned call_fsdriver_func_raw_rw(void __far* __far*, unsigned, unsigned long,
                                   void __far**, unsigned);
#pragma aux call_fsdriver_func_raw_rw = \
  "call far ptr es:[di]" \
  parm [es di] [bx] [dx ax] [si] [cx];

unsigned fsdriver_raw_read(unsigned long offset, void __far *buff, unsigned count, unsigned char drive)
{
  unsigned old_indos_flag, old_psp;
  unsigned rc;
  unsigned const driver_seg = partition_drvseg(drive);
  struct driver __far * const driver_ptr = MK_FP(driver_seg, 0);
  driver_func_entry(&old_indos_flag, &old_psp, driver_seg);
  rc = call_fsdriver_func_raw_rw(&driver_ptr->func_table[FS_RAWREAD], drive,
                                 offset, &buff, count);
  fsdriver_reterror(driver_seg, old_indos_flag, old_psp);
  return rc;
}

unsigned fsdriver_raw_write(unsigned long offset, void __far *buff, unsigned count, unsigned char drive)
{
  unsigned old_indos_flag, old_psp;
  unsigned rc;
  unsigned const driver_seg = partition_drvseg(drive);
  struct driver __far * const driver_ptr = MK_FP(driver_seg, 0);
  if(!driver_ptr->func_table[FS_RAWWRITE])
  {
    error_code = 1;
    return 0;
  }
  else
  {
    driver_func_entry(&old_indos_flag, &old_psp, driver_seg);
    rc = call_fsdriver_func_raw_rw(&driver_ptr->func_table[FS_RAWWRITE], drive,
                                   offset, &buff, count);
    fsdriver_reterror(driver_seg, old_indos_flag, old_psp);
    return rc;
  }
}
