#include "globals.c"
#include "macros.c"
#include "structs.c"

unsigned next_handle_seg(unsigned handle_seg)
{
  struct mcb __far *mcb = MK_FP(handle_seg - MCB_PARAGRAPHS, 0);

  while(mcb = MK_FP(mcb->next_mcb_seg, 0))
  {
    if(mcb->type == MCB_HANDLE)
      return FP_SEG(mcb) + MCB_PARAGRAPHS;
  }
  return 0;
}

// returns the segment of the handle struct or 0 if no such handle
unsigned find_handle(unsigned psp, unsigned handle)
{
  struct file_handle __far *handle_ptr = MK_FP(initial_mcb+MCB_PARAGRAPHS, 0);

  while(handle_ptr = MK_FP(next_handle_seg(FP_SEG(handle_ptr)), 0))
  {
    if(handle_ptr->owner_psp == psp && handle_ptr->id == handle)
      return FP_SEG(handle_ptr);
  }

  error_code = 0x06;

  return 0;
}

unsigned file_longest_handle(dirindex_t __far *file, unsigned long *length)
{
  unsigned retval = 0;
  struct file_handle __far *handle_ptr = MK_FP(initial_mcb+MCB_PARAGRAPHS, 0);
  unsigned char plen;
  *length = 0;
  for(plen=0;file[plen]!=-1;plen++);
  while(handle_ptr = MK_FP(next_handle_seg(FP_SEG(handle_ptr)), 0))
  {
    if(!memcmp(handle_ptr->file, file, plen*sizeof(dirindex_t)))
    {
      if(handle_ptr->length > *length) *length = handle_ptr->length;
      retval = FP_SEG(handle_ptr);
    }
  }

  return retval;
}

static void file_zero_current_clusters(struct file_handle __far *handle)
{
  struct file_handle __far *handle_ptr = MK_FP(initial_mcb+MCB_PARAGRAPHS, 0);

  while(handle_ptr = MK_FP(next_handle_seg(FP_SEG(handle_ptr)), 0))
    if(!memcmp(handle_ptr->file, handle->file, MAX_DIR_DEPTH+2)
       && FP_SEG(handle) != FP_SEG(handle_ptr))
    {
      memset(&handle_ptr->static_fss_info, 0, sizeof(union stat_fss_info));
      memcpy(&handle_ptr->file_fss_info, &handle->file_fss_info, sizeof(union file_fss_info));
      sync_dups(FP_SEG(handle_ptr));
    }
}

// finds free handle id for the program with the given owner_psp
// negative handle ids should be reserved as an error value
int find_free_handle(unsigned psp)
{
  unsigned retval = 0;

  while(find_handle(psp, retval))
    retval++;
  error_code = 0;
  
  return retval;
}

// allocates new handle for program
int new_handle(unsigned psp, unsigned type)
{
  int handle_id = find_free_handle(psp);
  unsigned handle_seg;
  struct file_handle __far *ptr;
  
  if(handle_id < 0) return -1; // no free handles

  handle_seg = allocate_high((sizeof(struct file_handle)+0xF)>>4, 0);
  
  if(!handle_seg) return -1; // out of memory

  set_mcb_type(handle_seg, MCB_HANDLE);
  
  ptr = MK_FP(handle_seg, 0);

  memset(ptr, 0, sizeof(struct file_handle));
  
  ptr->type = type;
  ptr->id = handle_id;
  ptr->owner_psp = psp;
  ptr->file[0] = -1;
  
  return handle_id;
}

// duplicate a file handle
// arguments: owner psp of the new handle, ID of the original handle, 
// requested id of the new handle (-1 if the first available)
// return -1 if failed, else return handle id
int dup_handle(unsigned psp, unsigned orig_seg, unsigned req_id)
{
  struct file_handle __far *orig_ptr = MK_FP(orig_seg, 0);
  struct file_handle __far *dup_ptr;
  unsigned dup_seg;
  int dup_id;

  if(delete_handle(psp, req_id, 0) == -2) // write error
    return -1;
  error_code = 0;

  dup_id = new_handle(psp, orig_ptr->type);
  if(dup_id < 0)
    return -1;
  dup_seg = find_handle(psp, dup_id);
  dup_ptr = MK_FP(dup_seg, 0);

  if(req_id != -1)
    dup_id = req_id;

  while(orig_ptr->dup)
    orig_ptr = MK_FP(orig_ptr->dup, 0);

  memcpy(dup_ptr, orig_ptr, sizeof(struct file_handle));
  orig_ptr->dup = dup_seg;
  dup_ptr->owner_psp = psp;
  dup_ptr->orig = orig_seg;
  dup_ptr->id = dup_id;

  return dup_id;
}

static void sync_dup(struct file_handle __far *dup_ptr, struct file_handle __far *orig_ptr)
{
  dup_ptr->length = orig_ptr->length;
  dup_ptr->offset = orig_ptr->offset;
  dup_ptr->write_buffer = orig_ptr->write_buffer;
  dup_ptr->buffer_size = orig_ptr->buffer_size;
  dup_ptr->is_written = orig_ptr->is_written;
  memcpy(&dup_ptr->file_fss_info, &orig_ptr->file_fss_info, sizeof(union file_fss_info));
  memcpy(&dup_ptr->static_fss_info, &orig_ptr->static_fss_info, sizeof(union stat_fss_info));
}

void sync_dups(unsigned handle_seg)
{
  struct file_handle __far *handle_ptr = MK_FP(handle_seg, 0);
  struct file_handle __far *dup_ptr = handle_ptr;

  while(dup_ptr = MK_FP(dup_ptr->orig, 0))
    sync_dup(dup_ptr, handle_ptr);

  dup_ptr = handle_ptr;
  while(dup_ptr = MK_FP(dup_ptr->dup, 0))
    sync_dup(dup_ptr, handle_ptr);
}

int delete_handle(unsigned psp, unsigned handle, unsigned char force)
{
  unsigned handle_seg = find_handle(psp, handle);
  struct file_handle __far *handle_ptr;
  struct file_handle __far *dup_ptr;
  
  if(!handle_seg)
    return -1;
  
  handle_ptr = MK_FP(handle_seg, 0);  

  if(handle_ptr->is_written)
  {
    if(flush_handle(handle_seg) == -1)
      if(!force)
      {
        write_error:
        return -2;
      }
      else
        free_buffers(handle_seg);

    if(update_file_info(handle_seg))
      if(!force)
        goto write_error;
  }

  if(!(handle_ptr->orig | handle_ptr->dup))
  {
    if(handle_ptr->type & 0x4000)
      driver_close(handle_ptr->file[0], handle_ptr->file[1]);
  }
  else
  {
    sync_dups(handle_seg);

    if(handle_ptr->orig)
    {
      dup_ptr = MK_FP(handle_ptr->orig, 0);
      dup_ptr->dup = handle_ptr->dup;
    }
    if(handle_ptr->dup)
    {
      dup_ptr = MK_FP(handle_ptr->dup, 0);
      dup_ptr->orig = handle_ptr->orig;
    }
  }

  delete_alloc(handle_seg);
  
  return 0;
}

void delete_psp_handles(unsigned psp)
{
  struct file_handle __far *handle_ptr = MK_FP(next_handle_seg(initial_mcb+MCB_PARAGRAPHS), 0);
  unsigned next_handle;

  do
  {
    next_handle = next_handle_seg(FP_SEG(handle_ptr));
    if(handle_ptr->owner_psp == psp)
      delete_handle(psp, handle_ptr->id, 1);
  }
  while(handle_ptr = MK_FP(next_handle, 0));
}

unsigned write_handle(unsigned psp, unsigned handle, unsigned char __far *data, unsigned count)
{
  unsigned handle_seg = find_handle(psp, handle);
  struct file_handle __far *handle_ptr;
  unsigned int retval = 0;
  
  if(!handle_seg) return 0;
  
  handle_ptr = MK_FP(handle_seg, 0);

  if(!handle_ptr->type) // console
  {
    write_con(data, count);
    return count;
  }
  if((handle_ptr->type & 3) == 1) // read only file
  {
    error_code = 0x05;
    return 0;
  }
  if(handle_ptr->type & 0x8000) // drive
  {
    if(count)
      retval = write_drive(handle_ptr->file[0], handle_ptr->offset, data, count);
    else
      retval = 0;
    handle_ptr->offset += retval;
  }
  else if(handle_ptr->type & 0x4000) // device
  {
    retval = driver_write(handle_ptr->file[0], handle_ptr->file[1], data, count);
  }
  else // file
  {
    if(!(handle_ptr->offset != handle_ptr->length
      && flush_handle(handle_seg) == -1))
    {
      if(count || handle_ptr->offset >= handle_ptr->length)
      {
        if(!count || write_to_buffer(handle_seg, count, data))
          retval = write_file(handle_ptr->offset, handle_ptr->file, count, data, handle_seg);
        else retval = count;
        handle_ptr->offset += retval;
        if(error_code == ERR_DISKFULL)
        {
          error_code = 0;
          if(!retval) goto end;
        }
        if(handle_ptr->offset > handle_ptr->length)
          handle_ptr->length = handle_ptr->offset;
      }
      else
      {
        if(retval = truncate_file(handle_seg))
          error_code = 0x0E;
        file_zero_current_clusters(handle_ptr);
        handle_ptr->length = handle_ptr->offset;
      }
      handle_ptr->is_written = 1;
    }
  }
  end:
  sync_dups(handle_seg);
  return retval;
}

static unsigned con_buff_getstr(unsigned count, char __far *buff)
{
  unsigned retval;

  retval = getstr(count-2, buff);
  buff[retval++] = 0x0D;
  buff[retval++] = 0x0A;
  return retval;
}

static unsigned read_con(unsigned char __far *buf, unsigned count)
{
  char __near *ptr;
  unsigned retval;
  if(ptr = (char __near*)strchr(CONINP_BUFF, 0x0A))
  {
    retval = ptr + 1 - CONINP_BUFF;
    read_from_buff:
    if(count < retval) retval = count;
    memcpy(buf, CONINP_BUFF, retval);
    memcpy(CONINP_BUFF, &CONINP_BUFF[retval], CONINPUT_BUFFER_SIZE-retval);
  }
  else
    if(count > 2)
      return con_buff_getstr(count, buf);
    else
    {
      retval = con_buff_getstr(CONINPUT_BUFFER_SIZE, CONINP_BUFF);
      goto read_from_buff;
    }

  return retval;
}

unsigned read_handle(unsigned psp, unsigned handle, unsigned char __far *data, unsigned count)
{
  unsigned handle_seg = find_handle(psp, handle);
  struct file_handle __far *handle_ptr;
  unsigned int retval;

  if(!handle_seg)  
    return 0;
  
  handle_ptr = MK_FP(handle_seg, 0);

  if(!handle_ptr->type) // console
  {
    return read_con(data, count);
  }
  if((handle_ptr->type & 3) == 2) // write only
  {
    error_code = 0x05;
    return 0;
  }
  if(handle_ptr->type & 0x4000) // device
  {
    retval = driver_read(handle_ptr->file[0], handle_ptr->file[1], data, count);
  }
  else if(handle_ptr->type & 0x8000) // drive
  {
    if(count)
      retval = read_drive(handle_ptr->file[0], handle_ptr->offset, data, count);
    else retval = 0;
  }
  else //file
  {
    if(handle_ptr->type == 3)
    {
      if(flush_handle(handle_seg) == -1) return 0;
    }
    retval = read_file(handle_ptr->offset, handle_ptr->file, count, data, handle_seg);
  }
  handle_ptr->offset += retval;
  sync_dups(handle_seg);
  return retval;
}
