#include "globals.c"
#include "macros.c"
#include "structs.c"

extern char fsdrivername[];

char __based(__segname("_CODE")) configsys[] = "CONFIG.SYS";

char __based(__segname("_CODE")) allocstack[] = "Allocated stack.\r\n";

char __based(__segname("_CODE")) searchcfg[] = "\r\nSearching for ";
char __based(__segname("_CODE")) notfound[] = "...not found\r\n";
char __based(__segname("_CODE")) cfgfound[] = "...found\r\n";
char __based(__segname("_CODE")) mounterror[] = "FS mount error ";
char __based(__segname("_CODE")) setvectors[] = "\r\nInterrupt vectors set.";

char __based(__segname("_CODE")) outofmemory[] = "Out of memory";

char __based(__segname("_CODE")) initializing_disks[] = "\r\nInitializing disks...\r\n";
char __based(__segname("_CODE")) initializing_fss[] = "Initializing filesystem driver ";
char __based(__segname("_CODE")) initializing_caches[] = "Initializing caches...";

char __based(__segname("_CODE")) benchmark_res[] = "Benchmark result: ";

char __based(__segname("_CODE")) starting_shell[] = "Loading shell.\r\n";

void int28h_handler();

#define LSHIFT_PRESSED ((*(unsigned char __far*)MK_FP(0x40, 0x17)) & 0x02)

static void set_vectors(void)
{
  set_near_vector(0x21, int21h_handler);
  set_near_vector(0x20, int20h_handler);
  set_near_vector(0x1B, int1bh_handler);
  set_near_vector(0x24, int24h_wrapper);
  set_near_vector(0x28, int28h_handler);
  set_near_vector(0x2F, int28h_handler);
  set_near_vector(0x2D, int28h_handler);
  set_near_vector(0x23, int23h_handler);
  if(init_timer_rtn())
  {
    putstr(outofmemory);
    halt();
  }
  putstr(setvectors);
}

void set_system_drive(void)
{
  if(SYSTEM_DRIVE & 0x80)
    WORKING_DRIVE = SYSTEM_DRIVE = FLOPPY_COUNT_R + (SYSTEM_DRIVE & 0x7F);
  else
    WORKING_DRIVE = SYSTEM_DRIVE;
}

extern unsigned char __far *buffer_segment;

static int alloc_work_buff(void)
{
  unsigned seg = new_alloc(MAX_SECTOR_SIZE>>4, 0);
  if(!seg) return 1;
  set_mcb_type(seg, MCB_FS);
  buffer_segment = MK_FP(seg, 0);
  return 0;
}

inline void copy_dpt(void)
{
  void __far *bios_dpt = get_vector(0x1e);
  memcpy(ALT_DPT, bios_dpt, 11);
  set_vector(0x1e, ALT_DPT);
}

void init_kernel_2(void)
{
  dirindex_t index;
  int rc;
  dirindex_t old_dir[MAX_DIR_DEPTH+2];

  putstr(initializing_caches);
  init_cache();

  if(memory_integrity_check())
    halt();

  putstr(initializing_disks);
  if(rc = initdisk())
    goto mount_error;
  if(memory_integrity_check())
    halt();

  putstr(initializing_fss);
  putstr(fsdrivername);
  if((rc = alloc_work_buff()) || (rc = init_fs_driver()))
  {
    mount_error:
    putstr(mounterror);
    putstr(hex2str(rc));
    halt();
  }
  if(memory_integrity_check())
    halt();

  set_vectors();

  copy_dpt();

  if(!LSHIFT_PRESSED)
  {
    putstr(searchcfg);
    putstr(configsys);


    index = get_file_index(SYSTEM_DRIVE, configsys, old_dir);
    if(!error_code)
    {
      putstr(cfgfound);
      parse_config(index);
    }
    else putstr(notfound);

    putstr(starting_shell);
  }
  memory_integrity_check();
}

