#include "globals.c"
#include "macros.c"
#include "structs.c"

static int allocate_program_resources(unsigned psp, unsigned parent)
{
  unsigned n;
  unsigned orig_seg;

  // stdin, stdout, stderr
  for(n=0;n<3;n++)
    if(parent && (orig_seg = find_handle(parent, n)))
    {
      if(dup_handle(psp, orig_seg, -1) == -1)
        return -1;
    }
    else
    {
      if(new_handle(psp, 0) < 0)
        return -1;
    }

  return 0;
}

static void free_program_resources(unsigned psp)
{
  delete_psp_handles(psp);
  free_psp_allocs(psp);
}

static void program_go_tsr(unsigned psp)
{
  unsigned paras = *(unsigned short __far*)MK_FP(psp, 2);
  unsigned char __far *ret_opcode = MK_FP(psp, 0);

  modify_alloc(psp, paras);
  *ret_opcode = 0xCD;
}

static unsigned load_com(dirindex_t *file, unsigned length, struct mzload *mz_struct)
{
  unsigned retval;
  unsigned char __far *code_buff;
  unsigned rc;
  unsigned min_paras;

  if(!mz_struct->load_overlay)
  {
    mz_struct->paras = free_paras(1, 1, 0x1000);

    if(mz_struct->paras < (min_paras = (length >> 4) + 0x10 + !!(length & 0xF)))
    {
      error_code = 0x08;
      mz_struct->paras = min_paras;
      return 0;
    }

    retval = new_alloc(mz_struct->paras, 1);
    code_buff = MK_FP(retval, 0x100);
  }
  else
    code_buff = MK_FP(mz_struct->initial_cs, 0);
  
  rc = read_file(0, file, 0xFFFF-0x100, code_buff, 0);
  if(error_code || rc != length)
  {
    if(!mz_struct->load_overlay)
      delete_alloc(retval);
    return 0;
  }
  if(!mz_struct->load_overlay)
  {
    mz_struct->initial_sp = (mz_struct->paras<<4) - 2;
    (*(unsigned __far*)MK_FP(retval, mz_struct->initial_sp)) = 0;
    mz_struct->initial_ip = 0x100;
    mz_struct->initial_ss = mz_struct->initial_cs = retval;
    return retval;
  }
  else
    return -1;
}

/*
 * Return values: 0 == error
   else returns program segment
 * */
unsigned load_program(char __far *program, char __far *args, unsigned env_seg, struct mzload *mz_struct)
{
  char __far *filename;
  dirindex_t index;
  struct file_info fileinfo;
  unsigned program_cs;
  dirindex_t file_index[MAX_DIR_DEPTH+3];
  dirindex_t old_dir[MAX_DIR_DEPTH+2];
  unsigned n;
  unsigned environment;
  unsigned t = 1;
  unsigned signature;
  int rc;

  flush_buffers();

  try_again:
  filename = chg2filedir(program, old_dir);
  if(error_code)
    return 0;

  if(!mz_struct->load_overlay)
  {
    environment = allocate_env_block(filename, env_seg);
    if(!environment) goto old_path;
  }

  index = get_file_index(WORKING_DRIVE, filename, old_dir);
  if(error_code)
  {
    old_path:
    chg2olddir(old_dir);
    errorexit:
    if(environment) delete_alloc(environment);
    if(error_code == 0x08 && t--)
    {
      if(make_more_memory(mz_struct->paras) >= mz_struct->paras)
      {
        error_code = 0;
        goto try_again;
      }
    }
    return 0;
  }

  fileinfo = file_info(WORKING_DRIVE, index);

  if(fileinfo.attr & 0x18) // if file is folder or volume label
  {
    error_code = 0x0D;
    goto old_path;
  }
  
  file_index[0] = WORKING_DRIVE; 
  n = getcwd(WORKING_DRIVE, &file_index[1]);

  chg2olddir(old_dir);

  file_index[n++] = index;
  file_index[n] = -1;

  rc = read_file(0, file_index, 2, &signature, 0);
  if(rc != 2 || (signature != 0x5A4D && signature != 0x4D5A))
  { // load a .COM program
    program_cs = load_com(file_index, fileinfo.size, mz_struct);
    if(!program_cs)
      goto errorexit;
  }
  else
  {
    program_cs = load_exe(file_index, mz_struct);
    if(!program_cs)
      goto errorexit;
  }

  if(!mz_struct->load_overlay)
  {
    set_mcb_owner(environment, program_cs);
    build_psp(program_cs, environment, mz_struct->paras, args);

    set_mcb_type(program_cs, MCB_PSP);
  }

  return program_cs;
}

inline void restore_interrupts(unsigned program_segment)
{

  set_vector(0x22, OLD_INT22_HDLR(program_segment));
  set_vector(0x23, OLD_BREAK_HDLR(program_segment));
  set_vector(0x24, OLD_CRITERR_HDLR(program_segment));
}

/*void load_debug(char __far *program, char __far *args, char __far *parm_block)
{
  struct mzload mz_struct;
  void __far * __far *stack = *(void __far**)&parm_block[0x0E];
  void __far * __far *jmpdst = *(void __far**)&parm_block[0x12];
  unsigned env_seg = *(unsigned __far*)&parm_block[0];

  load_program(program, args, env_seg, &mz_struct);

  *stack = MK_FP(mz_struct.initial_ss, mz_struct.initial_sp);
  *jmpdst = MK_FP(mz_struct.initial_cs, mz_struct.initial_ip);
}*/

void load_overlay(char __far *program, char __far *parm_block)
{
  struct mzload mz_struct;
  unsigned target_seg = *(unsigned __far*)&parm_block[0];
  unsigned reloc_factor = *(unsigned __far*)&parm_block[2];

  mz_struct.load_overlay = 1;
  mz_struct.initial_cs = target_seg;
  mz_struct.initial_ip = reloc_factor;

  load_program(program, 0, 0, &mz_struct);
}

int run_program(char __far *program, char __far *args, unsigned env_seg)
{
  struct mzload mz_struct;
  unsigned program_segment;

  mz_struct.load_overlay = 0;
  program_segment = load_program(program, args, env_seg, &mz_struct);
  if(!program_segment)
    return -1;

  if(allocate_program_resources(program_segment, CURRENT_PSP))
  {
    free_program_resources(program_segment);
    return -1;
  }

  CURRENT_PSP = program_segment;
  DTA_BUFFER = MK_FP(program_segment, 0x80);

  start_program(mz_struct.initial_cs, mz_struct.initial_ip,
              mz_struct.initial_ss, mz_struct.initial_sp);
}

unsigned end_program(void);
#pragma aux end_program value [ax];
unsigned end_program(void)
{
  unsigned program_segment = CURRENT_PSP;

  CURRENT_PSP = PARENT_PSP(program_segment); // parent psp
  if(CURRENT_PSP)
    TERMINATE_ADDRESS = get_vector(0x22);
  restore_interrupts(program_segment);

  if(!*(unsigned char __far*)MK_FP(program_segment, 0))
    program_go_tsr(program_segment);
  else
    free_program_resources(program_segment);

  return CURRENT_PSP;
}

