#include "globals.c"
#include "macros.c"
#include "structs.c"

unsigned char part2logicaldrive(unsigned char drive)
{
  struct partition __far *part = MK_FP(PARTITION_SEG[drive], 0);

  return part->logical_drive;
}

unsigned char part2physicaldrive(unsigned char drive)
{
  return logical2physicaldrive(part2logicaldrive(drive));
}

static lbaindex_t part2logicalsector(unsigned long sector, unsigned char drive)
{
  struct partition __far *part = MK_FP(PARTITION_SEG[drive], 0);

#ifdef USE_32BIT
  return b32_lba_addl(&part->d.starting_sector, &sector);
#else
  return part->d.starting_sector + sector;
#endif
}

static void set_prev_read_sector(unsigned long sector, unsigned char drive)
{
  unsigned char n;
  for(n=MAX_PARTITIONS;n--;)
    if(!partition_exists(n)/* || n == drive*/)
      continue;
    else
      PREV_READ_SECTOR(n) = -1;
  PREV_READ_SECTOR(drive) = sector;
}

extern unsigned char __far *buffer_segment;

static int img_write(unsigned long sector, struct partition __far *part, unsigned char drive)
{
  int rc;
  unsigned n;
  unsigned long offset;
  unsigned block_size = part->block_size;
  char rawbuff[MAX_SECTOR_SIZE];

  for(offset=sector,n=block_size;n>>=1;)
    offset <<= 1;

  memcpy(rawbuff, buffer_segment, block_size);
  rc = !(write_file(offset, part->file, block_size, rawbuff, 0) == block_size);
  memcpy(buffer_segment, rawbuff, block_size);
  part->is_written = 1;

  return rc;
}

int part_write(unsigned long sector, unsigned char drive)
{
  struct partition __far *part = MK_FP(PARTITION_SEG[drive], 0);
  int rc;

  if(part->file[0] != -1)
  {
    rc = img_write(sector, part, drive);
  }
  else
  {
    rc = write_logical_sector(part2logicalsector(sector, drive),
                              part2logicaldrive(drive));
  }
  set_prev_read_sector(sector, drive);

  return rc;
}

static int img_read(unsigned long sector, struct partition __far *part, unsigned char drive)
{
  int rc;
  unsigned n;
  unsigned long offset;
  unsigned block_size = part->block_size;
  char rawbuff[MAX_SECTOR_SIZE];

  for(offset=sector,n=block_size;n>>=1;)
    offset <<= 1;

  rc = !(read_file(offset, part->file, block_size, rawbuff, 0) == block_size);
  memcpy(buffer_segment, rawbuff, block_size);

  return rc;
}

int part_read(unsigned long sector, unsigned char drive, unsigned count)
{
  struct partition __far *part = MK_FP(PARTITION_SEG[drive], 0);
  int rc;

  if(PREV_READ_SECTOR(drive) == sector) return 0;

  if(sector + count >= part->size)
    count = part->size - sector;

  if(part->file[0] != -1)
  {
    rc = img_read(sector, part, drive);
  }
  else
  {
    rc = read_logical_sector(part2logicalsector(sector, drive), 
                             part2logicaldrive(drive), count);
  }
  set_prev_read_sector(sector, drive);

  return rc;
}

int add_partition(unsigned drive, lbaindex_t start_sector, unsigned long size, unsigned char type,
                  unsigned sector_size, unsigned char __far *drvname)
{
  unsigned n;
  struct partition __far *part_ptr;
  unsigned devseg, partnum, devnum;

  for(n=0;n<MAX_PARTITIONS;n++)
    if(!partition_exists(n))
      continue;
    else
    {
      part_ptr = MK_FP(PARTITION_SEG[n], 0);
      if(!(part_ptr->type & 0x80) && part_ptr->logical_drive == drive
         && part_ptr->d.starting_sector == start_sector)
        return n;
    }

  for(n=drive;PARTITION_SEG[n];)
    if(++n == MAX_PARTITIONS)
      return -1;

  PARTITION_SEG[n] = allocate_high((sizeof(struct partition)+0xF)>>4, 0);
  if(!PARTITION_SEG[n])
    return -1;

  set_mcb_type(PARTITION_SEG[n], MCB_SYSTEM);
  part_ptr = MK_FP(PARTITION_SEG[n], 0);
  memset(part_ptr, 0, sizeof(struct partition));

  if(!(type & 0x80))
  {
    part_ptr->logical_drive = drive;
    part_ptr->d.starting_sector = start_sector;
  }
  else
    if((devseg = is_driver_name(drvname, &partnum, &devnum))
       && devnum == -1)
    {
      part_ptr->d.driver_seg = devseg;
    }
    else
    {
      error_code = 0x03;
      delete_alloc(PARTITION_SEG[n]);
      return -1;
    }

  part_ptr->file[0] = -1;
  part_ptr->size = size;
  part_ptr->type = type;
  if(sector_size)
  {
    part_ptr->block_size = sector_size;
    SAVE_DRIVE_SECTOR_SIZE(drive, sector_size);
  }
  else
    part_ptr->block_size = MAX_SECTOR_SIZE;
  PREV_READ_SECTOR(n) = -1;

  return n;
}

int set_partition_size(unsigned char __far *func_tbl)
{
  const unsigned char drive = func_tbl[0];
  const unsigned sector_size = *(unsigned __far*)&func_tbl[2];
  const unsigned long size = *(unsigned long __far*)&func_tbl[4];
  struct partition __far * const part_ptr = MK_FP(PARTITION_SEG[drive], 0);
  if(!partition_exists(drive))
  {
    error_code = 0x0E;
    return -1;
  }

  part_ptr->block_size = sector_size;
  part_ptr->size = size;
  return 0;
}

int mount_drive(unsigned char __far *func_tbl, unsigned char __far *drvname)
{
  const unsigned char drive = func_tbl[0];
  const unsigned char type = func_tbl[1];
  const unsigned sector_size = *(unsigned __far*)&func_tbl[2];
  const unsigned long size = *(unsigned long __far*)&func_tbl[4];
  const lbaindex_t start_sector = *(lbaindex_t __far*)&func_tbl[8];
  const int rdrive = add_partition(drive, start_sector, size, type, sector_size, drvname);
  struct partition __far * const part_ptr = MK_FP(PARTITION_SEG[rdrive], 0);

  if(rdrive != -1 && (type & 0x80))
    part_ptr->ext = fsdriver_mount_partition(part_ptr->d.driver_seg, rdrive, 0);

  return rdrive;
}

int mount_file(unsigned char __far *func_tbl, unsigned char __far *drvname)
{
  struct partition __far *part_ptr;
  const unsigned char drive = func_tbl[0];
  const unsigned char type = func_tbl[1];
  const unsigned long size = *(unsigned long __far*)&func_tbl[2];
  dirindex_t old_dir[MAX_DIR_DEPTH+2];
  int rc;
  char __far *filename;
  dirindex_t index;
  int rdrive;

  filename = chg2filedir(&func_tbl[6], old_dir);
  if(!error_code)
  {
    index = get_file_index(WORKING_DRIVE, filename, old_dir);
    if(!error_code)
    {
      part_ptr = MK_FP(PARTITION_SEG[WORKING_DRIVE], 0);
      if(part_ptr->file[0] != -1)
        error_code = 0x0E;
      else
      {
        rdrive = add_partition(drive, 0, size, type, 0, drvname);
        if(rdrive != -1)
        {
          part_ptr = MK_FP(PARTITION_SEG[rdrive], 0);
          part_ptr->logical_drive = WORKING_DRIVE;
          part_ptr->file[0] = WORKING_DRIVE;
          rc = getcwd(WORKING_DRIVE, &part_ptr->file[1]);
          part_ptr->file[rc] = index;
          part_ptr->file[rc+1] = -1;
          if(part_ptr->type & 0x80)
            fsdriver_mount_partition(part_ptr->d.driver_seg, rdrive, &func_tbl[6]);
          else
            part_ptr->d.driver_seg = -1;
        }
      }
    }
  }
  chg2olddir(old_dir);

  return rdrive;
}

int unmount_partition(unsigned char partition, unsigned file)
{
  struct partition __far *part_ptr;
  int rc = 0;

  if(!partition_exists(partition))
    error_code = 0x0E;

  else
  {
    part_ptr = MK_FP(PARTITION_SEG[partition], 0);
    if(file)
    {
      if(part_ptr->file[0] == -1)
      {
        error:
        error_code = 0x0E;
        return 0;
      }
      else
        if(part_ptr->is_written)
          rc = update_file_info_nohandle(part_ptr->file);
    }
    else
    {
      if(part_ptr->file[0] != -1)
        goto error;
    }
    if(part_ptr->type & 0x80)
      fsdriver_unmount_partition(part_ptr->d.driver_seg, partition);
    else
      stfs_unmount_partition(partition);
    remove_partition(partition);
  }

  return rc;
}

void remove_partition(unsigned partition)
{
  delete_alloc(PARTITION_SEG[partition]);
  PARTITION_SEG[partition] = 0;
  if(WORKING_DRIVE == partition)
    WORKING_DRIVE = SYSTEM_DRIVE;
}
