#define MCB_NORMAL 0
#define MCB_DTA 1
#define MCB_CACHE 2
#define MCB_FS 3
#define MCB_HANDLE 4
#define MCB_SYSTEM 5
#define MCB_PSP 6
#define MCB_USERCACHE (MCB_CACHE | 0x80)

struct mcb
{
  unsigned next_mcb_seg;
  unsigned prev_mcb_seg;
  unsigned paragraph_count;
  unsigned owner_psp;
  unsigned char __far *usercache_freedptr;
  // 0: normal memory block
  // 1: dta struct
  // 2: disk cache
  // 3: filesystem
  // 4: handle
  // 5: system
  unsigned char type; 
};

struct chs
{
  unsigned cylinder;
  unsigned head;
  unsigned sector;
};

struct disk
{
  unsigned heads;
  unsigned char sectors;
};

struct time
{
  unsigned char ssecond:7;
  unsigned char second:6;
  unsigned char minute:6;
  unsigned char hour:5;
};

struct filetime
{
  unsigned twosec:5;
  unsigned minute:6;
  unsigned hour:5;
};

struct filedate
{
  unsigned day:5;
  unsigned month:4;
  unsigned year:7;
};

struct date
{
  unsigned year:11;
  unsigned char day:5;
  unsigned char month:4;
};

struct file_info
{
  unsigned long size;
  struct time time;
  struct date date;
  unsigned char attr;
};

// filesystem spesific information for file handle
union file_fss_info
{
  unsigned offset_cluster;
  long unused;
};
union stat_fss_info
{
  unsigned start_cluster;
  long unused;
};

struct file_handle
{
  union file_fss_info file_fss_info;
  union stat_fss_info static_fss_info;
  unsigned long length;
  unsigned long offset;
  dirindex_t file[MAX_DIR_DEPTH+3]; // drive + directory + file_index + -1
  unsigned buffer_size;
  int id;
  unsigned type;
  unsigned owner_psp;
  unsigned write_buffer;
  unsigned orig;
  unsigned dup;
  struct time time;
  struct date date;
  unsigned is_written:1;
  unsigned time_changed:1;
};

struct dta
{
  dirindex_t path[MAX_DIR_DEPTH+1];
  dirindex_t index;
  void __far *dta_ptr;
  unsigned id;
  char filename[FILENAME_LEN];
  unsigned char drive;
  unsigned char attr;
};

struct diskfree
{
  unsigned clusters;
  unsigned bytes_per_sector;
  unsigned unused_clusters;
  unsigned char sectors_per_cluster;
};

struct mzload
{
  unsigned initial_cs;
  unsigned initial_ip;
  unsigned initial_ss;
  unsigned initial_sp;
  unsigned paras;
  unsigned load_overlay:1;
};

struct cache_entry
{
  lbaindex_t sector;
  unsigned refs;
  unsigned norefs;
  unsigned buff_seg;
  unsigned length;
  unsigned char drive;
  unsigned valid:1;
  unsigned modified:1;
};

struct cache_retstr
{
  int index;
  unsigned char found[(RAW_CACHES<<4)/MAX_SECTOR_SIZE];
};

union partition_u
{
  lbaindex_t starting_sector;
  unsigned driver_seg;
};

#define PART_FAT12 1
#define PART_FAT16 0
struct partition
{
  dirindex_t file[MAX_DIR_DEPTH+3];
  union partition_u d;
  unsigned long size;
  unsigned long free_bytes;
  unsigned long buffer_bytes;
  unsigned long prev_read_sector;
  unsigned block_size;
  unsigned char logical_drive;
  unsigned char type;
  unsigned size_known:1;
  unsigned is_written:1;
  unsigned ext:1;
};

#define DRIVER_NAME_LEN 5
struct driver
{
  void __far * __far *func_table;
  unsigned next;
  unsigned psp;
  unsigned char name[DRIVER_NAME_LEN];
  unsigned type:1;
};

struct fw_buffer
{
  unsigned long offset;
  long seek_bytes;
  unsigned byte_count;
  unsigned alloc_bytes;
  unsigned handle_seg;
  unsigned buff_seg;
  unsigned next;
  union file_fss_info file_fss_info;
};
