#include <stdlib.h>
#include <stdio.h>
#include <i86.h>
#include <dos.h>

#define HARD_DISK_COUNT *(unsigned char __far*)MK_FP(0x40, 0x75)

unsigned char boot_drive;

int mount_partition(unsigned char, unsigned long, unsigned long);

int unmount_partition(unsigned char drive)
{
  char * const drvptr = &drive;
  int rc;

  printf("Unmounting partition %c\n", 'A' + drive);
  __asm
  {
    mov ah, 0x6a
    mov cx, 5
    mov dx, ds
    mov bx, drvptr
    int 0x21
    jnc noerror
    mov rc, ax
    jmp end
    noerror:
    mov rc, 0
    end:
  }

  return rc;
}

#define MSDOS_BOOT_INDICATOR(struct) struct[0]
#define MSDOS_BEGINNING_HEAD(struct) struct[1]
#define MSDOS_BEGINNING_SECTOR(struct) (struct[2]&0x3F)
#define MSDOS_BEGINNING_CYL(struct) ((((unsigned)struct[2] << 2) & 0x300) | struct[3])
#define MSDOS_SYSTEM_INDICATOR(struct) struct[4]
#define MSDOS_ENDING_HEAD(struct) struct[5]
#define MSDOS_ENDING_SECTOR(struct) (struct[6]&0x3F)
#define MSDOS_ENDING_CYL(struct) (((unsigned)struct[6] << 2) | struct[7])
#define MSDOS_HIDDEN_SECTORS(struct) (*(unsigned long*)&struct[8])
#define MSDOS_PARTITION_SIZE(struct) (*(unsigned long*)&struct[0xC])
/*
unsigned long logical_sector(unsigned char drive, unsigned cylinder, unsigned head, unsigned sector)
{
  unsigned char hd_heads, hd_sectors;
  unsigned long retval;
  unsigned char status;

  drive |= 0x80;

  __asm
  {
    mov ah, 0x08
    mov dl, drive
    int 0x13
    mov hd_sectors, cl
    inc dh
    mov hd_heads, dh
    mov ah, 0x01
    mov dl, drive
    int 0x13
    mov status, ah
  }

  if(status) return -1;

  if(sector)
    sector--;

  hd_sectors &= 0x3F;

  retval = cylinder * hd_heads * hd_sectors;
  retval += head * hd_sectors;
  retval += sector;

  return retval;
}
*/
int detect_msdospt(unsigned char drive, unsigned char root_drive, unsigned long logical_sector)
{
  static unsigned char buff[512];
  char filename[] = "DRV:A";
  FILE *fp;
  unsigned char *partstruct;
  unsigned n;
  unsigned long sector0[4];
  unsigned long sector1[4];
  unsigned char found = 0;
  unsigned char type[4];
  int rc;

  filename[4] = 'A'+drive;
  fp = fopen(filename, "rb");
  if(!fp) return 0;

  n = fread(buff, sizeof(buff), 1, fp);

  fclose(fp);

  if(n != 1) return 0;

  for(n=0,partstruct=buff+0x1BE;n<4;n++,partstruct+=0x10)
  {
    if(MSDOS_BOOT_INDICATOR(partstruct) & 0x7F)
      break;
/*    sector0[n] = logical_sector(drive, MSDOS_BEGINNING_CYL(partstruct),
       MSDOS_BEGINNING_HEAD(partstruct), MSDOS_BEGINNING_SECTOR(partstruct));
    sector1[n] = logical_sector(drive, MSDOS_ENDING_CYL(partstruct),
         MSDOS_ENDING_HEAD(partstruct), MSDOS_ENDING_SECTOR(partstruct));*/
    sector0[n] = MSDOS_HIDDEN_SECTORS(partstruct);
    sector1[n] = sector0[n] + MSDOS_PARTITION_SIZE(partstruct);
    if(sector0[n] && sector1[n] && sector0[n] >= sector1[n])
      break;
//    if(sector0[n] != MSDOS_HIDDEN_SECTORS(partstruct))
//      break;
//    if(sector1[n] - sector0[n] + 1 != MSDOS_PARTITION_SIZE(partstruct))
//      break;
    if(MSDOS_HIDDEN_SECTORS(partstruct))
      if(!MSDOS_PARTITION_SIZE(partstruct))
        break;
      else
      {
        found++;
        type[n] = MSDOS_SYSTEM_INDICATOR(partstruct);
      }
  }
  if(n != 4 || !found)
  {
    puts("No MSDOS partition table detected");
    return 0;
  }

  unmount_partition(drive);

  puts("MSDOS partition table detected, mounting partitions...");

  for(n=0;n<4;n++)
  {
    if(sector0[n])
    {
      sector0[n] += logical_sector;
      sector1[n] += logical_sector;
      printf("Partition %u starting sector: %lu, ending sector: %lu, partition\
 size %lu megabytes\n", n, sector0[n], sector1[n], 
        (sector1[n]-sector0[n])>>11);
      rc = mount_partition(root_drive, sector1[n] - sector0[n], sector0[n]);
      if(type[n] == 0x05 && rc != -1)
      {
        printf("Partition %c is a logical partition table - searching for more\
 partitions...\n", 'A'+rc);
        found += detect_msdospt(rc, root_drive, sector0[n]);
      }
    }
  }

  return found;
}

int mount_partition(unsigned char hard_drive, unsigned long size, unsigned long offset)
{
  unsigned char func_tbl[16];
  char * const tbl_ptr = func_tbl;
  int rc;
  int error;

  func_tbl[0] = hard_drive;
  func_tbl[1] = 0;
  *(unsigned*)&func_tbl[2] = 512;
  *(unsigned long*)&func_tbl[4] = size;
  *(unsigned long long*)&func_tbl[8] = offset;
  printf("Mounting partitions on device %u... ", hard_drive);
  __asm
  {
    mov di, tbl_ptr
    mov cx, 4
    mov ah, 0x6a
    int 0x21
    jc error1
    mov error, 0
    mov rc, ax
    jmp end
    error1:
    mov error, ax
    end:
  }
  if(!error)
    printf("mounted filesystem from offset %lu as partition %c!\n", offset, 'A' + rc);
  else
  {
    printf("Error %i\n", error);
    return -1;
  }

  return rc;
}

void mount_partitions(void)
{
  unsigned char hard_drive = 0;
  char func_tbl[10];
  int rc;
  unsigned char floppy_count = (*(unsigned char __far*)MK_FP(0x40, 0x10)) & 0xC1;
  if(floppy_count & 1)
  {
    floppy_count = ((floppy_count >> 6) + 1);
    if(floppy_count & 1)
      floppy_count++;
  }
  else
    floppy_count = 0;

//  if(boot_drive == 3) // drive C already mounted
//  if(boot_drive == floppy_count + 1)
//    hard_drive++;
//    unmount_partition(boot_drive-1);

  while(hard_drive < HARD_DISK_COUNT)
  {
    rc = mount_partition(hard_drive+floppy_count, -1, 0);
    if(rc != -1)
      detect_msdospt(rc, hard_drive+floppy_count, 0);
    hard_drive++;
  }
}

int main()
{
  puts("ST-DOS partition mounter v. 1.2");

  __asm
  {
    mov ax, 0x3305
    int 0x21
    mov boot_drive, dl
  }

  mount_partitions();

  return 0;
}
