#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "md5.h"

#ifdef DOS
#define NEWLINE_LEN 2
#else
#define NEWLINE_LEN 1
#endif

int calculate_md5sum(FILE *fp, char *md5sum)
{
  struct MD5Context md5struct;
  unsigned char read_buffer[512];
  unsigned char digest[16];
  unsigned char *digestp = digest;
  int rc;

  MD5Init(&md5struct);

  errno = 0;

  while(rc = fread(read_buffer, 1, sizeof(read_buffer), fp))
  {
    MD5Update(&md5struct, read_buffer, rc);
  }

  if(errno) return 1;

  MD5Final(digest, &md5struct);

  for(rc=0;rc<16;rc++)
  {
    sprintf(md5sum, "%.2X", *digestp++);
    md5sum += 2;
  }

  return 0;
}

int add_checksum(FILE *fp, char *md5sum)
{
  FILE *end_backup = fopen("_SECBOOT.TMP", "r");
  unsigned long fpos;
  int c;

  if(end_backup)
  {
    puts("File _SECBOOT.TMP already exists. Delete or rename it to generate \
checksum.");
    fclose(end_backup);
    return 1;
  }
  end_backup = fopen("_SECBOOT.TMP", "wt+");

  fseek(fp, -NEWLINE_LEN, SEEK_CUR);
  fpos = ftell(fp);

  while((c = fgetc(fp)) != EOF)
    if(fputc(c, end_backup) == EOF)
    {
      error:
      puts("Error while writing backup file");
      fclose(end_backup);
      return 2;
    }

  fseek(fp, fpos, SEEK_SET);

  if(fputc(' ', fp) == -1)
    goto error;
  if(fwrite(md5sum, 32, 1, fp) != 1)
    goto error;

  rewind(end_backup);

  while((c = fgetc(end_backup)) != EOF)
    if(fputc(c, fp) == EOF)
    {
      puts("Error while saving checksum. The database file is now corrupted, and you must fix it manually by using the \
contents of _SECBOOT.TMP file.");
      return 3;
    }

  fseek(fp, fpos, SEEK_SET);
  while(fgetc(fp) != '\n');

  fclose(end_backup);

  remove("_SECBOOT.TMP");

  return 0;
}

void check_file_integrity(FILE *fp)
{
  char line[200];
  char *end_of_filename;
  FILE *file;
  char md5sum[33];

  while(fgets(line, 200, fp))
  {
    end_of_filename = strchr(line, ' ');
    if(!end_of_filename)
      line[strlen(line)-1] = 0;
    else
      *end_of_filename = 0;
    file = fopen(line, "rb");
    if(!file)
    {
      printf("Could not open file %s. Press ENTER to continue", line);
      press_enter:
      getchar();
      continue;
    }
    if(calculate_md5sum(file, md5sum))
    {
      printf("Could not calculate the checksum of file %s. Press ENTER to \
continue", line);
      fclose(file);
      goto press_enter;
    }
    fclose(file);
    printf("md5sum of file %s: %s - ", line, md5sum);
    if(!end_of_filename)
    {
      printf("The md5sum of file %s is not yet saved to the database for comparing. Press ENTER to save the checksum \
and continue booting.", line);
      getchar();
      if(add_checksum(fp, md5sum)) goto fail;
    }
    else
    {
      end_of_filename++;
      if(memcmp(md5sum, end_of_filename, 32))
      {
        printf("Wrong checksum, should be %.32s! ", end_of_filename);
        fail:do
          puts("The system may be hacked or infected with a virus. Type \"Y\" and press ENTER to continue anyway.");
        while(getchar() != 'Y');
      }
      else
        puts("OK!");
    }
  }
}

int main(int argc, char *argv[])
{
  char *filename = argv[1];
  FILE *fp;

  if(argc < 2)
  {
    puts("Filelist file needed");
    puts("Example: >SECBOOT FILELIST.TXT");
    return 1;
  }

  fp = fopen(filename, "rt+");
  if(!fp)
  {
    printf("Could not open %s. Press enter to continue without security check", filename);
    getchar();
    return 2;
  }

  check_file_integrity(fp);

  return 0;
}
